#include "aggregation_ds.h"

using namespace Genode;

aggregation* create_aggregation(char aggn_type, int no_hash_types, ...)
{
	aggregation *aggn;
	aggn = new(Genode::env()->heap()) aggregation[1];
	aggn->aggregation_type = aggn_type;
	int i;
	va_list v1;
	va_start(v1, no_hash_types);

	//rather than calculate the size every time, calculate it here and store
	//the value.
	
	char *types = new(Genode::env()->heap()) char[no_hash_types + 1];
	va_start(v1, no_hash_types);
	for(i = 0;i < no_hash_types;i++)
	{
		types[i] = va_arg(v1, int);
	}
	va_end(v1);
	types[no_hash_types + 1] = '\0';
	aggn->hash_type_sequence = new(Genode::env()->heap()) char[no_hash_types + 1];
	memcpy(aggn->hash_type_sequence, types, strlen(types) + 1);

	destroy(Genode::env()->heap(), types);
	
	//allocating the memory for storing aggregation value data
	switch(aggn_type)
	{
		case SUM:
			//aggn->value_sequence = {DOUBLE, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[2];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = '\0';
			break;
		case MINIMUM:
			//aggn->value_sequence = {DOUBLE, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[2];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = '\0';
			break;
		case MAXIMUM:
			//aggn->value_sequence = {DOUBLE, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[2];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = '\0';			
			break;
		case COUNT:
			//aggn->value_sequence = {INT, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[2];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = '\0';
			break;
		case AVG:
			//aggn->value_sequence = {DOUBLE, DOUBLE, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[3];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = INT;
			aggn->value_sequence[2] = '\0';
			break;
		case STDEV:
			//aggn->value_sequence = {DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE, '\0'};
			aggn->value_sequence = new(Genode::env()->heap()) char[6];
			aggn->value_sequence[0] = INT;
			aggn->value_sequence[1] = INT;
			aggn->value_sequence[2] = INT;
			aggn->value_sequence[3] = INT;
			aggn->value_sequence[4] = INT;
			aggn->value_sequence[5] = '\0';;			
			break;
	}

	aggn->table = init_hash_table(aggn->hash_type_sequence, aggn->value_sequence);
	return aggn;
}

void sum_s(aggregation *aggn, char *hash_sequence, unsigned long value_add)
{
	//search the hash entry

	char *element = lookup_hash(aggn->table, hash_sequence);
	int val = value_add;

	//This does not exist, so initialize memory appropriately
	if(element == NULL)
	{
		insert_hash(aggn->table, hash_sequence, (char*)&val);
	}
	else //we have a pointer to the hash element, so just change the value
	{
		*((int*)element) += value_add;
	}
}

void count_s(aggregation *aggn, char *hash_sequence)
{
	char *element = lookup_hash(aggn->table, hash_sequence);
	//initialize memory if required
	
	if(element == NULL)
	{
		int one = 1;
		insert_hash(aggn->table, hash_sequence, (char*)&one);
	}
	else //otherwise, add 1 to value
	{
		*((int*)element) += 1;
	}
}

void min_s(aggregation *aggn, char *hash_sequence, int value)
{
	char *element = lookup_hash(aggn->table, hash_sequence);
	
	if(element == NULL)
	{
		insert_hash(aggn->table, hash_sequence, (char*)&value);
	}
	else if(value < *((int*)element))
	{
		*((int*)element) = value;
	}
}	

void max_s(aggregation *aggn, char *hash_sequence, int value)
{
	char *element = lookup_hash(aggn->table, hash_sequence);
	
	if(element == NULL)
	{
		insert_hash(aggn->table, hash_sequence, (char*)&value);
	}
	else if(value > *((int*)element))
	{
		*((int*)element) = value;
	}
	//initialize if required (initial: 0);
	//check: (>), assign as appropriate
}

void stdev_s(aggregation *aggn, char *hash_sequence, int value)
{
	//check correct type
	//the number of elements for stdev needs to be examined further
	//initialize elements as necessary
}

void avg_s(aggregation *aggn, char *hash_sequence, int value)
{
	char *element = lookup_hash(aggn->table, hash_sequence);

	if(element == NULL)
	{
		//2 elements require initialisation: cumulative total, and count
		double values[2] = {value, 1};
		insert_hash(aggn->table, hash_sequence, (char*)values);
	}
	else
	{
		//add as necessary
		*((int*)element) += value;
		*((int*)element + sizeof(double)) += 1;
	}
}

int sum_w(buffer_info *info, char *values)
{
	/*
		Need to make sure we have something to consolidate over
		ie. [int, int, ..., INT] ie. the last type must be an int.
	*/
	int sum_total = 0;
	char *value_ptr = values;
	if(info->storage_mode == CONSOLIDATE)
	{
		while(*value_ptr != WINDOW_END)
		{
			//the aggregation value should immediately follow the hash sequence.
			int key_value = value_size(info, value_ptr);
			sum_total += *((int*)(value_ptr + key_value));
			value_ptr += value_size(info, value_ptr) + sizeof(int);
		}
	}
	else
	{
		return -1;
	}
	
	return sum_total;
}

int avg_w(buffer_info *info, char *values)
{
	/*
		Need to make sure we have something to consolidate over
		ie. [int, int, ..., INT] ie. the last type must be an int.
	*/
	int sum_total = 0;
	int count = 0;
	char *value_ptr = values;
	if(info->storage_mode == CONSOLIDATE)
	{
		while(*value_ptr != WINDOW_END)
		{
			//the aggregation value should immediately follow the hash sequence.
			int key_value = value_size(info, value_ptr);
			sum_total += *((int*)(value_ptr + key_value));
			value_ptr += value_size(info, value_ptr) + sizeof(int);
			count++;
		}
	}
	else
	{
		return -1;
	}
	
	if(count != 0)
		return sum_total/count;
	
	return -1;
}

int count_w(buffer_info *info, char *values)
{
	/*
		Need to make sure we have something to consolidate over
		ie. [int, int, ..., INT] ie. the last type must be an int.
	*/
	char *value_ptr = values;
	int sum_total = 0;
	if(info->storage_mode == CONSOLIDATE)
	{
		while(*value_ptr != WINDOW_END)
		{
			//the aggregation value should immediately follow the hash sequence.
			sum_total++;
			value_ptr += value_size(info, value_ptr) + sizeof(int);
		}
	}
	else
	{
		return -1;
	}
	
	return sum_total;
}

char *min_w(buffer_info *info, char *values)
{
	char *value_ptr = values;
	char *current_min_data = new(Genode::env()->heap()) char[value_size(info, value_ptr) + sizeof(int)];
	int ptr_current_min = value_size(info, value_ptr);
	
	value_ptr += value_size(info, value_ptr) + sizeof(int);
	memcpy(current_min_data, value_ptr, value_size(info, value_ptr) + sizeof(int));
	
	if(info->storage_mode == CONSOLIDATE)
	{
		while(*value_ptr != WINDOW_END)
		{
			int key_value = value_size(info, value_ptr);
			
			//checking whether the appropriate aggregation value is < the current aggregation value
			if(*((int*)(value_ptr + key_value)) < *((int*)(current_min_data + ptr_current_min)))
			{
				destroy (Genode::env()->heap(), current_min_data);
				current_min_data = new(Genode::env()->heap()) char[value_size(info, value_ptr) + sizeof(int)];
				memcpy(current_min_data, value_ptr, key_value + sizeof(int));
				value_ptr += key_value + sizeof(int);
			}
		}
	}
	
	return current_min_data;
}

char *convert_string_hash(aggregation *aggn, int no_args, ...)
{
	struct hash_table *table = aggn->table;
	char *data_sequence = NULL;
	char *next_argument = NULL;

	char type_value = 0;
	int initial_run = 1;
	int bytes_written = 0;

	va_list v1;
	va_start(v1, no_args);
	
	unsigned int i;
	for(i = 0;i < strlen(table->hash_types);i++)
	{	
		type_value = table->hash_types[i];
		
		switch(type_value)
		{
			case INT:
				{
					unsigned long initial = va_arg(v1, unsigned long);
					//printf("value of iteration is %d\n", i);
					//printf("value of long is %ld\n", initial);
					if(initial < 0)
					{
						printf("NEGATIVE VALUES NOT ALLOWED\n");
						return NULL;
						//initial *= -1;
					}
					
					if(initial_run == 1)
					{//malloc an initial size for the integer
						data_sequence = new(Genode::env()->heap()) char[sizeof(int)];
						bytes_written += sizeof(int);
						*((unsigned long*)data_sequence) = initial;
						initial_run = 0;
					}
					else
					{
						data_sequence = realloc(data_sequence, bytes_written, bytes_written + sizeof(int));
						
						//append the new data
						next_argument = data_sequence + bytes_written;
						*((unsigned long*)next_argument) = initial;
						bytes_written += sizeof(int);
						//printf("done\n");
					}
				}
				break;
			case DOUBLE:
				{
					double initial = va_arg(v1, double);
					
					if(initial_run == 1)
					{
						data_sequence = new(Genode::env()->heap()) char[sizeof(double)];
						bytes_written += sizeof(double);
						*((double*)data_sequence) = initial;
						initial_run = 0;
					}
					else
					{
						//create the additional memory for the data sequence.  THIS HAS TO BE CHECKED VERY
						//CAREFULLY GIVEN realloc() COULD EXHIBIT UNPREDICTABLE BEHAVIOUR.  POSSIBLE BUG!!!
						data_sequence = realloc(data_sequence, bytes_written, bytes_written + sizeof(double));
						
						next_argument = data_sequence + bytes_written;
						*((double*)next_argument) = initial;
						bytes_written += sizeof(double);
					}
				}
				break;
			case STRING:
				{
					char *initial = va_arg(v1, char*);
					if(initial_run == 1)
					{
						data_sequence = new(Genode::env()->heap()) char[strlen(initial) + 1];
						bytes_written += strlen(initial);
						memcpy((void*)data_sequence, (void*)initial, bytes_written);
						*((char*)(data_sequence + bytes_written + 1)) = '\0';
						bytes_written++;
						initial_run = 0;
					}
					else
					{
						//create the additional memory for the data sequence.  THIS HAS TO BE CHECKED VERY
						//CAREFULLY GIVEN realloc() COULD EXHIBIT UNPREDICTABLE BEHAVIOUR.  POSSIBLE BUG!!!
						data_sequence = realloc(data_sequence, bytes_written, bytes_written + strlen(initial) + 1);
						
						next_argument = data_sequence + bytes_written;
						memcpy((void*)next_argument, (void*)initial, strlen(initial));
						bytes_written += strlen(initial);
						*((char*)(next_argument + strlen(initial) + 1)) = '\0';
						bytes_written++;
					}
				}
				break;
			/*case INSTN:
				{
					unsigned long initial = va_arg(v1, unsigned long);
					if(initial_run == 1)
					{
						data_sequence = (char*)malloc(sizeof(unsigned long));
						bytes_written += sizeof(unsigned long);
						*((unsigned long*)data_sequence) = initial;
						initial_run = 0;
					}
					else
					{
						//create the additional memory for the data sequence.  THIS HAS TO BE CHECKED VERY
						//CAREFULLY GIVEN realloc() COULD EXHIBIT UNPREDICTABLE BEHAVIOUR.  POSSIBLE BUG!!!
						data_sequence = (char*)realloc(data_sequence, bytes_written + sizeof(unsigned long));
						
						next_argument = data_sequence + bytes_written;
						*((unsigned long*)next_argument) = initial;
						bytes_written += sizeof(unsigned long);
					}
				}
				break;*/
		}
		
	}
	va_end(v1);
	
	return data_sequence;
}
char *max_w(buffer_info *info, char *values)
{
	char *value_ptr = values;
	char *current_max_data = new(Genode::env()->heap()) char[value_size(info, value_ptr) + sizeof(int)];
	int ptr_current_max = value_size(info, value_ptr);
	
	value_ptr += value_size(info, value_ptr) + sizeof(int);
	memcpy(current_max_data, value_ptr, value_size(info, value_ptr) + sizeof(int));
	
	if(info->storage_mode == CONSOLIDATE)
	{
		while(*value_ptr != WINDOW_END)
		{
			int key_value = value_size(info, value_ptr);
			
			//checking whether the appropriate aggregation value is < the current aggregation value
			if(*((int*)(value_ptr + key_value)) > *((int*)(current_max_data + ptr_current_max)))
			{
				destroy(Genode::env()->heap(), current_max_data);
				current_max_data = new(Genode::env()->heap()) char[value_size(info, value_ptr) + sizeof(int)];
				memcpy(current_max_data, value_ptr, key_value + sizeof(int));
				value_ptr += key_value + sizeof(int);
			}
		}
	}
	
	return current_max_data;
}

int stdev_w(buffer_info *info, char *values)
{
	return 0;
}

int value_size(buffer_info *info, char *values)
{
	char *hash_seq = (char*)info->type_information;
	
	int total = 0;
	
	unsigned int i;
	char *iter = values;
	for(i = 0;i < strlen(hash_seq);i++)
	{
		switch(hash_seq[i])
		{
			case INT:
				total += sizeof(int);
				iter += sizeof(int);
				break;
			case DOUBLE:
				total += sizeof(double);
				iter += sizeof(double);
				break;
			case STRING:
			{
				int length = strlen(iter);
				total += length + 1;
				iter += length + 1;
				break;
			}
			case BOOLEAN:
				total += 1;
				iter += 1;
				break;
		}
	}
	
	return total;
}

struct hash_entry *first_entry(aggregation *aggn)
{
	struct hash_table *table_ptr = aggn->table;
	return &table_ptr->entry1;
}
