#include "data_output.h"
#include <base/printf.h>
#include <util/string.h>

using namespace Genode;

void int_output(L4_ThreadId_t tid, int data)
{
	buffer_info *info = thread_buffer_info(tid.raw);
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		info->type_identifier = -1;
		initialize_buffer(NULL, info, NULL, tid);
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index(L4_Myself()));
	}
	
	write_output(info, &data);
	
}

void double_output(buffer_info *info, double data, L4_ThreadId_t tid)
{
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		info->type_identifier = -2;
		initialize_buffer(NULL, info, NULL, tid);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index(L4_Myself()));
	}
	
	write_output(info, &data);	
}

void string_output(buffer_info *info, char *data, L4_ThreadId_t tid)
{
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		info->type_identifier = -3;
		initialize_buffer(NULL, info, NULL, tid);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index(L4_Myself()));
	}
	
	write_output(info, &data);	
}

void boolean_output(buffer_info *info, char data, L4_ThreadId_t tid)
{
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		info->type_identifier = -4;
		initialize_buffer(NULL, info, NULL, tid);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index(L4_Myself()));
	}
	
	write_output(info, &data);	
}

void tuple_output(buffer_info *info, void *data, L4_ThreadId_t tid)
{
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		//type identifier is always the first field of a tuple/struct
		info->type_identifier = *((int*)data);
		initialize_buffer(NULL, info, NULL, tid);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		int given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index(L4_Myself()));
	}
	
	write_output(info, data);	
}

void aggregation_output(L4_ThreadId_t writer, aggregation *data)
{
	buffer_info *info = binfo_raw_thread(writer.raw);
	//buffer_info *info = binfo_fn((uintptr_t)fn);
	int hash_length = strlen(data->hash_type_sequence);

	//type sequence for aggregation: hash type sequence + double type
	char *sequence = new(Genode::env()->heap()) char[hash_length + 2];
	char *seq_ptr = sequence;
	memcpy((void*)seq_ptr, data->hash_type_sequence, hash_length);
	seq_ptr += hash_length;
	char dble = DOUBLE;
	memcpy((void*)seq_ptr, &dble, 1);
	seq_ptr++;
	*((char*)seq_ptr) = '\0';

	info->type_identifier = find_type_seq(sequence);

	tinfo *info_about_type = get_stats(info->type_identifier);
	info->actual_chunk_size = memory_single_chunk((char*)info_about_type->tup_type_seq) + sizeof(int);//additional id field
	
	int counter;
    struct hash_entry *entry;
    struct hash_table *table = data->table;
    int hash_mem_size = memory_single_chunk(data->hash_type_sequence);
 	char *data_to_output = new(Genode::env()->heap()) char[hash_mem_size + 12];
 	char *data_ptr = data_to_output;
    
	for (counter = 0; counter < 48; counter++) 
    {
        entry = &table->entries[counter];
        //printf("value of counter is %d\n", counter);
        while (entry) 
        {
        	if(entry->hash_sequence != NULL)
        	{
        		int type = info->type_identifier;
        		memcpy(data_ptr, &type, sizeof(int));
        		data_ptr += sizeof(int);
        		memcpy(data_ptr, entry->hash_sequence, hash_mem_size);
        		data_ptr += hash_mem_size;
        		
        		switch(data->aggregation_type)
        		{
        			case MAXIMUM:
        			case MINIMUM:
        			case COUNT:
        			case SUM:
        			{
						memcpy(data_ptr, entry->value_sequence, sizeof(double));
						break;
        			}
        			case AVG:
        			{
        				double sum = *((double*)entry->value_sequence);
        				double divisor = *((double*)(entry->value_sequence + sizeof(double)));
        				double quotient = sum / divisor;
        				memcpy(data_ptr, &quotient, sizeof(double));
        				break;
        			}
        			case STDEV:
        			{
        				double temp_value = 0;
        				memcpy(data_ptr, &temp_value, sizeof(double));
        				break;
        			}
        			
        		}
    			//print_hash_key(table, entry);
        		//printf("\t\t\t%ld\n", *((long*)entry->value_sequence));
        		write_output(info, data_to_output);
        		data_ptr = data_to_output;
        	}
	        
            entry = entry->next;
        }
    }
}

typedef struct
{
	int id;
	int value1;
	int value2;
	double value3;
}temp_tup;

void write_output(buffer_info *info, void *data)
{
	L4_Word_t first_time = 0;
	if(info->reset == 0)
	{
		if((uintptr_t)info->write_index + info->actual_chunk_size + 1 < (uintptr_t)info->buffer + info->size)
		{
			memcpy((void*)info->write_index, data, info->actual_chunk_size);
			//printf("the value just written is %d\n", *((int*)info->write_index));
			info->write_index += info->actual_chunk_size;
		}
		else
		{
			uintptr_t min_index = minimum_index_bounds(info);
			
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS && 
				(uintptr_t)info->buffer + info->actual_chunk_size + 1 < (uintptr_t)info->buffer + min_index)
			{
				first_time = 1;
			}
			else
			{
				printf("user records: We have problems\n");
				L4_Stop_Thread(L4_Myself());
			}
			
			info->reset = 1;
		}
	}
	
	if(info->reset == 1)
	{
		uintptr_t min_index = minimum_index_bounds(info);
		
		//Reset the write index pointer in the case where there are threads links that have
		//read some data from the buffer.
		if(first_time == 1 && min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS)
		{
			first_time = 0;
			*(info->write_index) = BUFFER_RESET;
			info->write_index = info->buffer;
		}
		
		//In the case where there are either no links or that no threads have scanned the buffer,
		//we just drop the data.
		if(min_index == NO_THREAD_LINKS || min_index == NO_BUFFER_SCANS)
		{
			info->dropped_samples++;
		}
		else if((uintptr_t)info->write_index + info->actual_chunk_size + 1 < (uintptr_t)info->buffer + min_index)
		{//Otherwise, see if we can fit the data in the remaining space.  Given we are copying
		 //constant sized chunks, this is unlikely to be the case.
			/*if(first_time == 1)
			{
				info->write_index += sizeof(L4_Word_t) * 2;
			}*/
			
			memcpy((void*)info->write_index, data, info->actual_chunk_size);
			info->write_index += info->actual_chunk_size;
			
		}
		else
		{
			info->dropped_samples++;
		}
	}
}
