#include "profile_port.h"
#include "my_threads.h"

static sampling_buffer_t *samples;
//static Genode::network_buffer *net_ptr;

//Genode::network_buffer *network_buf;

L4_ThreadId_t interrupt_id;
L4_ThreadId_t net_interrupt_id;

void resume_prof_threads()
{
	int counter;
	
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	L4_ThreadId_t tid;
	
	for(counter = 0;counter < matrix_index;counter++)
	{
		tid.raw = thread_list[counter];
		L4_Start(tid);
	}	
}

void netsamp_thread::thread_fn()
{
	net_interrupt_id.raw = tid().l4id.raw;
	printf("NETWORK ID is %ld\n", net_interrupt_id.raw);
//	net_ptr = network_buf;
	interrupt_spec intrpt;
	intrpt.handler_id = net_interrupt_id;
	//Genode::network_profile(HANDLER_ID, &intrpt);
	
	while(1)
	{
		L4_MsgTag_t tag;
		L4_ThreadId_t sender;
        tag = L4_Wait(&sender);
        
        L4_Word_t finished = samples->finished;
        
        if(finished != 1)
        {
        
        }
        else
        {
        }
	}
}

void netsamp_thread::entry()
{
	thread_fn();
}

void prof_interrupt_handler::thread_fn()
{
	enum{SAMP_PHYS = 0x440000, SAMP_SIZE = 0x40000};
	Io_mem_connection sampbuffer(SAMP_PHYS, SAMP_SIZE);
	samples = (sampling_buffer_t*)env()->rm_session()->attach(sampbuffer.dataspace());
	sampbuffer.on_destruction(Io_mem_connection::KEEP_OPEN);
	if(samples == NULL)
		printf("THE BUFFER SHOULD NOT BE NULL\n");

	interrupt_id.raw = tid().l4id.raw;
	interrupt_spec intrpt;
    intrpt.handler_id = interrupt_id;
    L4_ProfPolicy(HANDLER_ID, &intrpt);

	while(1)
	{
		L4_Stop_Thread(interrupt_id);
		L4_Word_t finished = samples->finished;

		if(finished != 1)
		{		
			copy_sample_contents();
		}
		else
		{
			copy_sample_contents_finish();
			samples->finished = 0;
			finished = samples->finished;
		}
		resume_prof_threads();
	}
}

void prof_interrupt_handler::entry()
{
	thread_fn();
}

L4_Word_t get_reset_state(buffer_info *info, L4_Word_t index)
{
	if(info->threads[index] != NULL)
	{
		return info->resetted[index];
	}
	
	return -1;
}

L4_Word_t all_threads_resetted(buffer_info *info)
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{
		if(info->threads[i] != NULL && info->resetted[i] == 1)
		{
			return 0;
		}
	}
	
	return 1;
}

void get_write_index(buffer_info *info, L4_Word_t *address)
{
	write_index_operation(info, READ, address, 0);
}

void change_write_index(buffer_info *info, L4_Word_t amount)
{
	write_index_operation(info, WRITE, 0, amount);
}

void write_index_operation(buffer_info *info, char operation, L4_Word_t *addr, L4_Word_t amount)
{
	if(operation == READ)
		*addr = (uintptr_t)info->write_index;
	else
		info->write_index += amount;
		
}

char *obtain_raw_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
							int *no_record)
{
	*no_record = 0;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	L4_Word_t current_offset = thread_buffer->t_offsets[source_thread_index];

	char *current_index = (char*)(thread_buffer->buffer + current_offset);
	L4_Word_t current_index_addr = (uintptr_t)current_index;
	L4_Word_t wrt_index;
	get_write_index(thread_buffer, &wrt_index);
	L4_Word_t thread_reset_state = get_reset_state(thread_buffer, source_thread_index);
	char *new_buffer = NULL;

	if(thread_reset_state == 0)
	{
		if(current_index == NULL)
		{
			printf("current index is NULL\n");
			L4_Stop_Thread(L4_Myself());
		}
		else if(current_index_addr >= wrt_index)
		{
			*no_record = 1;
		}
		else if(Genode::memcmp(current_index, "T_END", 5) == 0 || Genode::memcmp(current_index, "S_END", 5) == 0 ||
				Genode::memcmp(current_index, "M_END", 5) == 0 || Genode::memcmp(current_index, "H_END", 5) == 0 ||
				Genode::memcmp(current_index, "END", 4) == 0)
		{//It must either be one of the end particular sampling tags (eg. SEC_END) or 
		 //a profiling tag.  A PROF_END tag should imply that collection has been finalized for 
		 //the given thread anyway.
			new_buffer = (char*)"END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 6;
			*no_record = 1;
			return new_buffer;
		}
		else if(memcmp(current_index, "P1_END", 6) == 0 || memcmp(current_index, "P2_END", 6) == 0)
		{
			new_buffer = (char*)"END";
			thread_buffer->t_offsets[source_thread_index] += 7;
			*no_record = 1;
			return new_buffer;
		}
		else
		{
			//here, check that we can get the sample record.  If not, we switch to any thread.  Otherwise, obtain
			//data as normal.
			if(current_index_addr + thread_buffer->actual_chunk_size >= wrt_index)
			{
				*no_record = 1;
			}
			else
			{
				char *new_buffer = new(env()->heap()) char[thread_buffer->actual_chunk_size];
				Genode::memcpy(new_buffer, (void*)current_index, thread_buffer->actual_chunk_size);
				thread_buffer->t_offsets[source_thread_index] += thread_buffer->actual_chunk_size;//L4sample_size(current_index);
				thread_buffer->moved[source_thread_index] = 1;

				return new_buffer;
			}
		}
	}
	
	if(thread_reset_state == 1)
	{
		if(*current_index == BUFFER_RESET)
		{//Upon encountering this tag, the read index is resetted.  If all threads have resetted,
		 //the buffer state is set to non-reset ie. samples are copied to the buffer as per normal.
			thread_buffer->t_offsets[source_thread_index] = 0;
			thread_buffer->no_scanned++;
			thread_buffer->moved[source_thread_index] = 0;
			thread_buffer->resetted[source_thread_index] = 0;
			//If all threads have scanned through the buffer, sample writing can 
			//take place as normal.
			if(all_threads_resetted(thread_buffer) == 1)
			{
				thread_buffer->no_scanned = 0;
				thread_buffer->reset = 0;
			}
			*no_record = 1;
		}
		else
		{
			char *new_buffer = new(env()->heap())char[thread_buffer->actual_chunk_size];
			memcpy(new_buffer, (void*)current_index, thread_buffer->actual_chunk_size);
			thread_buffer->t_offsets[source_thread_index] += thread_buffer->actual_chunk_size;
			thread_buffer->moved[source_thread_index] = 1;
			return new_buffer;
		}
	}
	
	new_buffer = new(env()->heap()) char[1];
	*new_buffer = 'z';
	return new_buffer;
}

char *obtain_sample_window(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
							int *no_record)
{
	*no_record = 0;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	L4_Word_t current_offset = thread_buffer->t_offsets[source_thread_index];
	char *current_index = (char*)(thread_buffer->buffer + current_offset);
	L4_Word_t current_index_addr = (uintptr_t)current_index;

	L4_Word_t wrt_index;
	get_write_index(thread_buffer, &wrt_index);
	L4_Word_t thread_reset_state = get_reset_state(thread_buffer, source_thread_index);
	
	L4_Word_t window_size = determine_size(thread_buffer) / thread_buffer->no_windows + 1;
	
	char *new_buffer;
	
	if(thread_reset_state == 0)
	{
		if(current_index == NULL)
		{
			printf("obtain_sample_window: This thread is not permitted to access the specified buffer\n");
			L4_Stop_Thread(L4_Myself());
		}
		else if(current_index_addr >= wrt_index)
		{
			*no_record = 1;
		}
		/*else if(*current_index == MSEC_END || *current_index == SEC_END ||
				*current_index == MIN_END || *current_index == HR_END ||
				*current_index == PROF_END)
		{
			new_buffer = "END";
			thread_buffer->t_offsets[source_thread_index]++;
			*no_record = 1;
			return new_buffer;
		}*/
		else if(memcmp(current_index, "T_END", 5) == 0 || memcmp(current_index, "S_END", 5) == 0 ||
				memcmp(current_index, "M_END", 5) == 0 || memcmp(current_index, "H_END", 5) == 0)
		{//It must either be one of the end particular sampling tags (eg. SEC_END) or 
		 //a profiling tag.  A PROF_END tag should imply that collection has been finalized for 
		 //the given thread anyway.
			const char *new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 6;
			*no_record = 1;
			return (char*)new_buffer;
		}
		else if(memcmp(current_index, "P1_END", 6) == 0 || memcmp(current_index, "P2_END", 6) == 0)
		{
			const char *new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 7;
			*no_record = 1;
			return (char*)new_buffer;
		}
		else
		{
			if(current_index_addr + window_size > wrt_index)
			{
				*no_record = 1;
			}
			else
			{
				new_buffer = new(env()->heap())char[window_size];
				memcpy((void*)new_buffer, (void*)current_index, window_size);
				thread_buffer->t_offsets[source_thread_index] += window_size;
				thread_buffer->moved[source_thread_index] = 1;
				return new_buffer;
			}
		}
	}
	
	if(thread_reset_state == 1)
	{
		if(*current_index == BUFFER_RESET)
		{
			thread_buffer->t_offsets[source_thread_index] = 0;
			thread_buffer->no_scanned++;
			thread_buffer->moved[source_thread_index] = 0;
			
			thread_buffer->resetted[source_thread_index] = 0;
			
			if(all_threads_resetted(thread_buffer) == 1)
			{
				thread_buffer->no_scanned = 0;
				thread_buffer->reset = 0;
			}
			
			*no_record = 1;
			//current_index = (char*)(thread_buffer->buffer);
		}
		else if(*current_index == MSEC_END || *current_index == SEC_END ||
				*current_index == MIN_END || *current_index == HR_END ||
				*current_index == PROF_END)
		{
			const char *new_buffer = "END";
			thread_buffer->t_offsets[source_thread_index]++;
			*no_record = 1;
			return (char*)new_buffer;
		}
		else
		{
			new_buffer = new(env()->heap())char[window_size];
			memcpy((void*)new_buffer, (void*)current_index, window_size);
			thread_buffer->t_offsets[source_thread_index] += window_size;
			thread_buffer->moved[source_thread_index] = 1;
			return new_buffer;
		}
	}
	
	new_buffer = new(env()->heap())char[1];
	*new_buffer = 'z';
	return new_buffer;
}

char *obtain_sample_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record)
{
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	
	if(thread_buffer->storage_mode == NO_CONSOLIDATE)
		return obtain_raw_record(destination_thread_index, source_thread_index, no_record);

	return obtain_sample_window(destination_thread_index, source_thread_index, no_record);

}

//char *retrieve_data(void (*fn1)(void*), void (*fn2)(void*), int *no_record)
//retrieve_data(collection1, processing1, &nothing)
char *retrieve_data(buffer_info *info, L4_ThreadId_t my_id, int *no_record)
{
	//buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	//L4_Word_t input_thread_index = thread_ptr_index((uintptr_t)fn1);
	//L4_Word_t source_thread_index = thread_ptr_index((uintptr_t)fn2);//thread_index(current_thread);
	L4_Word_t input_thread_index = buffer_info_index(info);
	L4_Word_t source_thread_index = thread_index(my_id);
	
	return obtain_sample_record(input_thread_index, source_thread_index, no_record);
}

void copy_raw_record(char *sample_record, buffer_info *dest_thread_buffer)
{
	L4_Word_t first_time = 0;
	L4_Word_t current_write_ptr = (uintptr_t)dest_thread_buffer->write_index;
	L4_Word_t buffer_address = (uintptr_t)dest_thread_buffer->buffer;
	L4_Word_t min_index = minimum_index_bounds(dest_thread_buffer);
	
	L4_Word_t sample_chunk_size = dest_thread_buffer->actual_chunk_size;
	
	if(dest_thread_buffer->reset == 0)
	{
		if(current_write_ptr + sample_chunk_size >= buffer_address + dest_thread_buffer->size)
		{
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS &&
				buffer_address + sample_chunk_size  + dest_thread_buffer->connected_threads < buffer_address + min_index)
			{
				first_time = 1;
			}

			set_reset_state(dest_thread_buffer);
			dest_thread_buffer->reset = 1;
		}
		else
		{
			//Write type id information
			*((int*)dest_thread_buffer->write_index) = dest_thread_buffer->type_identifier;
			dest_thread_buffer->write_index += sizeof(int);

			//TID information if specified
			if(dest_thread_buffer->info_collect & INFO_TID)
			{
				int threadid = L4samp_get_tid(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->write_index, (void*)&threadid, sizeof(int));
				dest_thread_buffer->write_index += sizeof(int);
			}
			
			//instruction address information if specified
			if(dest_thread_buffer->info_collect & INFO_INST)
			{
				unsigned int instn = L4samp_get_inst(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->write_index, &instn, sizeof(unsigned int));
				dest_thread_buffer->write_index += sizeof(unsigned int);
			}
		}
	}
	
	if(dest_thread_buffer->reset == 1)
	{
		if(first_time == 1)
		{
			first_time = 0;
			*(dest_thread_buffer->write_index) = BUFFER_RESET;
			dest_thread_buffer->write_index = dest_thread_buffer->buffer;
			current_write_ptr = (uintptr_t)dest_thread_buffer->write_index;
		}
		else if(current_write_ptr + dest_thread_buffer->actual_chunk_size < buffer_address + min_index)
		{
			*((int*)dest_thread_buffer->write_index) = dest_thread_buffer->type_identifier;
			dest_thread_buffer->write_index += sizeof(int);	
					
			if(dest_thread_buffer->info_collect & INFO_TID)
			{
				int threadid = L4samp_get_tid(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->write_index, (void*)&threadid, sizeof(int));
				dest_thread_buffer->write_index += sizeof(int);
			}
			
			if(dest_thread_buffer->info_collect & INFO_INST)
			{
				unsigned int instn = L4samp_get_inst(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->write_index, &instn, sizeof(unsigned int));
				dest_thread_buffer->write_index += sizeof(unsigned int);
			}
		}
		else
		{
//			printf("about to drop a sample\n");
			dest_thread_buffer->dropped_samples++;
		}
	}
}

L4_Word_t insert_aggregation(buffer_info *info, char *record)//, char *buffer, char *current_ptr, char *record, char agg_code)
{
	/*
		For the given record, scan through the consolidation buffer. From the current scan pointer,
		scan through the buffer.  If we get beyond the write pointer of the consolidated buffer,
		add the sample to the buffer along with the corresponding value.  Otherwise, increment
		the value for the given sample.
	*/
	
	char *entry = (char*)info->intermediate_buffer;
	while((uintptr_t)entry < (uintptr_t)info->intermediate_ptr)
	{//incrementing current_ptr???? NO -> should only happen if the sample has not been found
		//entry -> intermediate buffer. Hence, need to do entry +4
		if(sample_equals(info, record, (char*)(entry + 4)) == 1)
		{
			//increment the count value appropriate (COUNT is the only operation
			//that is possible for consolidation of samples)
			if(info->info_collect & INFO_TID && info->info_collect & INFO_INST)
			{
				*((int*)(entry + 12)) += 1;
			}
			else if(info->info_collect & INFO_TID || info->info_collect & INFO_INST)
			{
				*((int*)(entry + 8)) += 1;
			}
			
			return 1;
		}
		
		entry += info->actual_chunk_size;
	}

	Genode::size_t sample_size = info->actual_chunk_size - 8;
	*((int*)entry) = info->type_identifier;
	entry += sizeof(int);
	Genode::memcpy((void*)entry, (void*)record, sample_size);
	entry += sample_size;
	*((int*)entry) = 1;
	info->intermediate_ptr += info->actual_chunk_size;
	
	return 0;
}

void copy_core_samples_over(buffer_info *info)
{	
	char *buffer_ptr = (char*)info->core_samples;
	L4_Word_t found;
	//every sample requires an additional chunk of memory storing the count value
	int consol_size = determine_size(info);
	int window_size = consol_size / info->no_windows + sizeof(char);
	info->intermediate_buffer = new(Genode::env()->heap()) char[window_size];//(char*)malloc(window_size);
	info->intermediate_ptr = info->intermediate_buffer;

	while((uintptr_t)buffer_ptr < (uintptr_t)info->core_pointer)
	{//i should actually copy the samples into an intermediate buffer and after that, THEN
	 //copy the data into the buffer_info buffer
	 	//allocate memory for the temporary buffer.  The amount to allocate is the same as
	 	//the size of the core buffer.

		found = insert_aggregation(info, buffer_ptr);

		buffer_ptr += info->actual_chunk_size - 8;
	}

	*(info->intermediate_ptr) = WINDOW_END;
	info->intermediate_ptr++;

	//copy the intermediate buffer (sorting the now consolidated data)
	//into the sampling buffer
	Genode::memcpy((void*)info->write_index, (void*)info->intermediate_buffer, 
					(Genode::size_t)((uintptr_t)info->intermediate_ptr - (uintptr_t)info->intermediate_buffer));
	
	info->core_pointer = info->core_samples;
	info->write_index += window_size;

	destroy(Genode::env()->heap(), (char*)info->intermediate_buffer);
	info->intermediate_ptr = NULL;
}

void copy_sample_window(char *sample_record, buffer_info *dest_thread_buffer)
{//at this point, i need to create records storing the sample data AND an integer for count
	L4_Word_t first_time = 0;
	L4_Word_t size_record = dest_thread_buffer->actual_chunk_size;//L4sample_size(sample_record);
	L4_Word_t current_write_ptr = (uintptr_t)dest_thread_buffer->write_index;
	L4_Word_t buffer_address = (uintptr_t)dest_thread_buffer->buffer;
	L4_Word_t min_index = minimum_index_bounds(dest_thread_buffer);
	L4_Word_t current_core_ptr = (uintptr_t)dest_thread_buffer->core_pointer;
	L4_Word_t core_address = (uintptr_t)dest_thread_buffer->core_samples;
	
	L4_Word_t window_size = determine_size(dest_thread_buffer) /
							dest_thread_buffer->no_windows + sizeof(char);
	
	if(dest_thread_buffer->reset == 0)
	{
		if(current_write_ptr + window_size >= buffer_address + dest_thread_buffer->size)
		{
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS &&
				buffer_address + window_size < buffer_address + min_index)
			{
				first_time = 1;
			}
			set_reset_state(dest_thread_buffer);
			dest_thread_buffer->reset = 1;
		}
		else
		{
			if(current_core_ptr + size_record - 8 < core_address + dest_thread_buffer->core_size)
			{
				if(dest_thread_buffer->info_collect & INFO_TID)
				{
					int threadid = L4samp_get_tid(sample_record);
					Genode::memcpy((void*)dest_thread_buffer->core_pointer, (void*)&threadid, sizeof(int));
					dest_thread_buffer->core_pointer += sizeof(int);
					current_core_ptr += sizeof(int);
				}
				
				if(dest_thread_buffer->info_collect & INFO_INST)
				{
					unsigned int instn = L4samp_get_inst(sample_record);
					Genode::memcpy((void*)dest_thread_buffer->core_pointer, &instn, sizeof(unsigned int));
					dest_thread_buffer->core_pointer += sizeof(unsigned int);
					current_core_ptr += sizeof(int);
				}
			}
			
			if(current_core_ptr >= core_address + dest_thread_buffer->core_size - 1)
			{
				copy_core_samples_over(dest_thread_buffer);
			}
		}
	}
	
	if(dest_thread_buffer->reset == 1)
	{
		if(first_time == 1)
		{
			first_time = 0;
			*(dest_thread_buffer->write_index) = BUFFER_RESET;
			dest_thread_buffer->write_index = dest_thread_buffer->buffer;
			current_write_ptr = (uintptr_t)dest_thread_buffer->write_index;
		}
		else if(current_core_ptr + size_record - 8 < core_address + dest_thread_buffer->core_size)
		{
			if(dest_thread_buffer->info_collect & INFO_TID)
			{
				int threadid = L4samp_get_tid(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->core_pointer, (void*)&threadid, sizeof(int));
				dest_thread_buffer->core_pointer += sizeof(int);
				//current_core_ptr += sizeof(int);
			}
			
			if(dest_thread_buffer->info_collect & INFO_INST)
			{
				unsigned int instn = L4samp_get_inst(sample_record);
				Genode::memcpy((void*)dest_thread_buffer->core_pointer, &instn, sizeof(unsigned int));
				dest_thread_buffer->core_pointer += sizeof(unsigned int);
			//	current_core_ptr = sizeof(int);
			}
		}
		else if(current_core_ptr >= core_address + dest_thread_buffer->core_size - 1)
		{
			if(current_write_ptr + window_size < buffer_address + min_index)
				copy_core_samples_over(dest_thread_buffer);
			else
				dest_thread_buffer->dropped_samples++;
		}
	}
}

/* sample_record MUST point to the current sampling buffer entry */
void copy_sample_record(char *sample_record, buffer_info *dest_thread_buffer)
{//re-examine: probably don't need 2nd param given the sample record will tell us where to record the information

	if(dest_thread_buffer->storage_mode == NO_CONSOLIDATE)
		copy_raw_record(sample_record, dest_thread_buffer);
	else
		copy_sample_window(sample_record, dest_thread_buffer);
}

void copy_sample_contents()
{
	uintptr_t buffer, count;
	uintptr_t buffer_mask;
	
	//obtain head information....
	buffer = samples->active_buffer;
	buffer_mask = samples->buffer_empty;
	
	count = samples->buffers;
	
	while(count--)
	{
		if((buffer != samples->active_buffer) && ((~buffer_mask) & (1ul << buffer)))
		{
			//printf("Copying...\n");
			
			//this is where we copy the buffer elements character by character.
			//This should be fast given it's character by character.
			uintptr_t head = samples->buffer_head[buffer];
			uintptr_t entry = 0;
			char *current_entry = (char*)samples + samples->buffer_offset[buffer] + entry;
			/* 	
				This loop should terminate when the current pointer entry has 
				exceeded the current write index.  I must verify what value 'buffer_head' 
				refers to. It should pertain to the current index that data will be 
				written to.  There cannot be an 'end of profiling' tag as that should
				mean that the other copying function should be invoked.
			*/
			while(entry < head)
			{
				buffer_info *info;
				current_entry = (char*)samples + samples->buffer_offset[buffer] + entry;
				//printf("Current entry value is %ld\t\t head value is %ld\n", (uintptr_t)current_entry, head);
	        	//printf("We are here\n");
			
            	//if this is not true, the tag must represent the start of a sample.
            	if(*current_entry == PROF_START)
            	{
            		entry++;
            	}
            	else if(*((L4_Word_t*)current_entry) == CALLSTACK_END)
            	{
            		//printf("We are at the end of the callstack\n");
            		entry += sizeof(L4_Word_t);
            	}
                else if(*current_entry == MSEC_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_MSEC))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_MSEC)));
            		}
            		
            		info = thread_buffer_info(tunit_index(TID_MSEC));
            		//*(thread_buffer_info(tunit_index(TID_MSEC))->write_index) = MSEC_END;
            		Genode::memcpy((void*)info->write_index, "T_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
   				else if(*current_entry == SEC_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_SEC))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_SEC)));
            		}
            		
            		info = thread_buffer_info(tunit_index(TID_SEC));
            		//*(thread_buffer_info(tunit_index(TID_SEC))->write_index) = SEC_END;
            		Genode::memcpy((void*)info->write_index, "S_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
            	else if(*current_entry == MIN_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_MIN))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_MIN)));
            		}
					
					info = thread_buffer_info(tunit_index(TID_MIN));
            		//*(thread_buffer_info(tunit_index(TID_MIN))->write_index) = MIN_END;
            		Genode::memcpy((void*)info->write_index, "M_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
                else if(*current_entry == HR_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_HR))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_HR)));
            		}

					info = thread_buffer_info(tunit_index(TID_HR));
            		//*(thread_buffer_info(tunit_index(TID_HR))->write_index) = HR_END;
            		Genode::memcpy((void*)info->write_index, "H_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
            	else if(*current_entry == PROF_END)
            	{//If this is encountered, there is the possibility that more samples may be obtained,
            	 //depending on how the developer specifies the threads to execute, so just increment
            	 //the pointer in the case that this is true.
            		//printf("I have a PROF END record\n");
            		entry++;
            	}
                else if(*current_entry == SAMPLE_START)
                {
            	  	//Make sure i can get the sample size record.  If i can't, it must be
          		  	//in the other buffer, which is dealt with by the other conditions.
          		  	L4_Word_t size = sizeof(char) + sizeof(L4_Word_t);
          		  	if(entry + size < head && (entry + L4sample_size(current_entry) < head))
          		  	{
          		  		L4_Word_t raw_id = *((L4_Word_t*)(current_entry + sizeof(char)));
						copy_sample_record(current_entry, thread_buffer_info(raw_id));
	               		entry += L4sample_size(current_entry);
          		  	}
                }
				else
				{
					PDBG("copy sample contents: I SHOULD NOT APPEAR\n");
					while(1);
				}
			}//while(entry < head)
	
			samples->buffer_empty |= (1ul << buffer);
			//printf("Copying has finished\n");
		
		}
		buffer++;
		//We've only got 2 buffers, so reset it to index 0 for the
		//first element
		if(buffer >= samples->buffers)
			buffer = 0;
	}
}

void copy_sample_contents_finish()
{
    uintptr_t buffer, count;
    uintptr_t buffer_mask;

    buffer = samples->active_buffer;
    buffer_mask = samples->buffer_empty;

    if(buffer == ~(0ul))
            buffer = 0;
	int number_samples = 0;
    count = samples->buffers;
    
    while(count--)
    {
        if(buffer == samples->active_buffer) 
        {	
			//uintptr_t head = samples->buffer_head[buffer];
			uintptr_t entry = 0;
			L4_Word_t raw_id;

			char *current_entry = (char*)samples + samples->buffer_offset[buffer] + entry; 


			/*
				This loop MUST terminate when it encounters an 'PROF_END' tag.  Otherwise,
				the other function should be invoked, not this one.
			*/
			
            while(*current_entry != PROF_END)
            {
        		current_entry = (char*)samples + samples->buffer_offset[buffer] + entry;
            	buffer_info *info;

            	//if this is not true, the tag must represent the start of a sample.
            	if(*current_entry == PROF_START)
            	{
            		entry++;
            	}
            	else if(*((L4_Word_t*)current_entry) == CALLSTACK_END)
            	{
            		entry += sizeof(L4_Word_t);
            	}
                else if(*current_entry == MSEC_END)
            	{//Maybe it might be better to have the MSEC_END in the window itself rather than
            	 //separately in another window.  This would mean that in terms of consolidating, a case
            	 //should be provided where i cover end of sampling tags.
            		if(thread_buffer_info(tunit_index(TID_MSEC))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_MSEC)));
            		}

					info = thread_buffer_info(tunit_index(TID_MSEC));

            		//*(thread_buffer_info(tunit_index(TID_MSEC))->write_index) = MSEC_END;
            		Genode::memcpy((void*)info->write_index, "T_END", 6);
					
            		info->write_index += 6;

            		entry++;
            	}
   				else if(*current_entry == SEC_END)
            	{
	           		if(thread_buffer_info(tunit_index(TID_SEC))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_SEC)));
            		}

					info = thread_buffer_info(tunit_index(TID_SEC));
            		//*(thread_buffer_info(tunit_index(TID_SEC))->write_index) = SEC_END;
            		Genode::memcpy((void*)info->write_index, "S_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
            	else if(*current_entry == MIN_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_MIN))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_MIN)));
            		}

					info = thread_buffer_info(tunit_index(TID_MIN));
            		//*(thread_buffer_info(tunit_index(TID_MIN))->write_index) = MIN_END;
            		Genode::memcpy((void*)info->write_index, "M_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
                else if(*current_entry == HR_END)
            	{
            		if(thread_buffer_info(tunit_index(TID_HR))->storage_mode == CONSOLIDATE)
            		{
            			copy_core_samples_over(thread_buffer_info(tunit_index(TID_HR)));
            		}	            	
					
					info = thread_buffer_info(tunit_index(TID_HR));
            		//*(thread_buffer_info(tunit_index(TID_HR))->write_index) = HR_END;
            		Genode::memcpy((void*)info->write_index, "H_END", 6);
            		info->write_index += 6;
            		entry++;
            	}
            	//else if(*current_entry == NSAMP_END)
            	//{
            	//	if(thread_buffer_info(tunit_index(TID_HTTP))->storage_mode == CONSOLIDATE)
            	//	{
            	//		copy_core_samples_over(thread_buffer_info(tunit_index(TID_HTTP)));
            	//	}
            		
            	//	info = thread_buffer_info(tunit_index(TID_HR));
            		//*(thread_buffer_info(tunit_index(TID_HR))->write_index) = HR_END;
            	//	Genode::memcpy((void*)info->write_index, "H_END", 6);
            	//	info->write_index += 6;
            	//	entry++;
            		
            	//}        	            	
               	else if(*current_entry == PROF_END)
               	{
               		entry++;
            		break;
                }
                else if(*current_entry == SAMPLE_START)
                {
          		  	raw_id = *((L4_Word_t*)(current_entry + sizeof(char)));
					copy_sample_record(current_entry, thread_buffer_info(raw_id));
                	entry += *((L4_Word_t*)(current_entry + sizeof(char) + sizeof(L4_Word_t)));
                	number_samples++;
                }
                else
                {
                	PDBG("copy sample contents finish: I SHOULD NOT APPEAR\n");
                	while(1);
                }
            }
            
   			//Don't worry about switching buffers at this point, just reset the pointers
   			samples->buffer_head[buffer] = 0;
        }
        
        buffer++;
        
        //We've only got 2 buffers, so reset it to index 0 for the
        //first element
        if(buffer >= samples->buffers)
        	buffer = 0;
    }
    printf("Number of samples collected is %d\n", number_samples);
}
