namespace Okl4 { extern "C" {
#include <l4/arch/thread.h>
#include <l4/types.h>
#include <l4/thread.h>
#include <l4/ipc.h>
#include <l4/arch/types.h>
#include <l4/kdebug.h>
#include <l4/profpolicy.h>
}}

extern "C" {
#include "tracebuffer.h"
#include <stddef.h>
}

#include <base/printf.h>
#include <base/env.h>
#include <base/sleep.h>
#include <base/thread.h>
#include <util/string.h>
#include <cpu_session/cpu_session.h>
#include <dataspace/client.h>
#include <io_mem_session/connection.h>

#include "type_info.h"
#include "prof_vector.h"
//#include "my_hash.h"

#define SAMPLING_THREAD 11244
#define THREAD_QUANTITY 100

#define TID_MSEC	0
#define TID_SEC		1
#define TID_MIN		2
#define TID_HR		3
#define TID_PMC1	4
#define TID_PMC2	5
#define TID_HTTP	6

#define NO_THREAD_LINKS	0
#define NO_BUFFER_SCANS 1

#define NO_USER_GEN_RECORDS	0
#define AGG_SAMPLES -1
#define USER_RECORDS -2

#define RAW		'0'
#define AGGN	'1'
#define TUPLE	'2'

#define END_OF_PERIOD		'Y'
#define END_OF_RECORDS		'Z'
#define NO_CONSOLIDATE	0
#define CONSOLIDATE		1
#define WINDOW_END '0'

#define READ '0'
#define WRITE '1'

#define NO_RECORD_WRITES	0
#define NORMAL_WRITE		1
#define BUFFER_RESETTED		2

#define NO_TYPE	'\0'
#define INT		'I'
#define DOUBLE	'D'
#define STRING	'S'
#define BOOLEAN 'B'

typedef unsigned long int	uintptr_t;

using namespace Okl4;

typedef struct
{
	int volatile interrupt_no;
	int volatile info_collect;
	int volatile thread_index;
	int volatile duration;
	int volatile rate;
	int volatile time_window;
	int volatile window_unit;
	int volatile no_windows;
	int volatile storage_mode;

	char volatile buffer_type;
	
	/* Buffer to store raw samples for a fixed time length.  Only used where
		samples are being consolidated. */
	char volatile *core_samples;
	
	/* The current write index of the raw sampling buffer */
	char volatile *core_pointer;
	
	/* Size of the core buffer */
	L4_Word_t volatile core_size;
	
	/* Stores the relevant type information for the stream*/
	char volatile *type_information;
	
	/* Type information stored in stream. Either a basic type (in which information is  */
	tinfo volatile *stream_type;
	int volatile type_identifier;
	
	char volatile *intermediate_buffer;
	char volatile *intermediate_ptr;

	/* The current buffer writing index */ 
	char volatile *write_index;
	
	/* The sampling buffer (which gets malloc'ed).  In consolidation mode,
		this is used as the consolidated buffer.  Otherwise, this is used
		as the buffer for storing raw samples. */
	char volatile *buffer;
	
	/* An array to keep track of the threads inspecting the buffer */
 	char volatile *threads[THREAD_QUANTITY];
 	
 	/* Offsets of threads connected to the buffer */
 	L4_Word_t volatile t_offsets[THREAD_QUANTITY];
 	
 	/* A simple array that determines whether an inspecting thread has
 	   started scanning or not */
 	int volatile moved[THREAD_QUANTITY];
 	
 	/* The size of the sampling buffer */
	L4_Word_t size;
	
	/* State variable to determine whether the write index has gone back to
		the beginning of the buffer */
	L4_Word_t volatile reset;
	
	/* Maintain current reset states of all sampling threads */
	L4_Word_t volatile resetted[THREAD_QUANTITY];
		
	/* The number of threads linked to this particular buffer (besides the 
		thread that the buffer is linked to */
	int volatile connected_threads;
	
	/* For all the connected threads, how many of them have scanned through the buffer */
	L4_Word_t volatile no_scanned;
	
	/* Counter to count the number of samples dropped due to insufficient space */
	L4_Word_t volatile dropped_samples;
	
	/* For writing and reading from the buffer purely in user space, the chunk size
	   determines the size of the record to be obtained from the buffer */
	Genode::size_t volatile chunk_size;
	
	Genode::size_t volatile actual_chunk_size;
	
	Genode::size_t volatile remaining_size;
}buffer_info;


#define FIRST_SAMPLE(buffer)	buffer + sizeof(char) * 2 + sizeof(L4_Word_t) * 6
#define SECOND_SAMPLE(buffer)	FIRST_SAMPLE(buffer) + sizeof(char) + sizeof(L4_Word_t)
#define THIRD_SAMPLE(buffer)	SECOND_SAMPLE(buffer) + sizeof(char) + sizeof(L4_Word_t)

//Obtaining individual elements: Use the reference to the sample itself, not
//the struct storing this information.  The thread's read index should be modified
//when the whole sample record is obtained beforehand.
L4_INLINE int L4samp_get_tid(char *buffer)
{
	//Make sure that there is actually a TID record.  The TID record does not
	//exist in the case where the scanning of the sample type immediately
	//results in the record for the current instruction address
	if(*(FIRST_SAMPLE(buffer)) == SAMPLE_TID)
	{
		return *((int*)(FIRST_SAMPLE(buffer) + sizeof(char)));
	}
	else
	{
		return -1;
	}
}

//Get the information relating to the current instruction that was executing
L4_INLINE L4_Word_t L4samp_get_inst(char *buffer)
{
	//Make sure there is actually an instruction record.  In this case,
	//increment up to the first tag.  If the tag is an instruction tag, then get the
	//record.
	if(*(FIRST_SAMPLE(buffer)) == INSTN_ADDR)
	{
		//get the information
		return *(FIRST_SAMPLE(buffer) + sizeof(char));
	}
	else if(*(SECOND_SAMPLE(buffer)) == INSTN_ADDR)
	{
		//get the information
		return *(SECOND_SAMPLE(buffer) + sizeof(char));
	}
	else
	{
		return -1;
	}
}

L4_INLINE L4_Word_t L4sample_size(char *buffer)
{
	return *((L4_Word_t*)(buffer + sizeof(char) + sizeof(L4_Word_t)));
}


extern volatile int matrix_index;
extern volatile int thread_matrix[THREAD_QUANTITY][THREAD_QUANTITY];
extern volatile buffer_info *thread_buffers[THREAD_QUANTITY];

extern volatile L4_ThreadId_t thread_tunit[7];
extern L4_Word_t thread_list[THREAD_QUANTITY];
extern L4_ThreadId_t list_of_reg_threads[THREAD_QUANTITY];

extern uintptr_t thread_ptrs[THREAD_QUANTITY];
extern volatile L4_Word_t thread_list_index;

extern L4_ThreadId_t register_id;
extern L4_ThreadId_t matrix_id;
extern L4_ThreadId_t interrupt_id;

void assert(bool cond);
char* realloc(char *old_buffer, int old_size, int new_size);

void init_matrix();

uintptr_t minimum_index_bounds(buffer_info *buffer);
void set_reset_state(buffer_info *info);
L4_Word_t determine_size(buffer_info *info);
L4_Word_t sample_equals(buffer_info *info, char *sample1, char *sample2);
L4_Word_t tunit_index(int tunit);
buffer_info* thread_buffer_info(int index);

L4_Word_t enter_thread_entry(L4_Word_t tid_register, buffer_info *buffer);//uintptr_t fn_ptr, buffer_info *buffer);
void enter_matrix_entry(int source_thread_index, int destination_thread_index);

void copy_sample_contents();
void copy_sample_contents_finish();
void resume_prof_threads();
buffer_info* binfo_fn(uintptr_t fn);
int memory_single_chunk(char *sequence);

void init_threads();

void initialize_binfo(buffer_info *info);
int thread_index(L4_ThreadId_t input);
void initialize_buffer(interrupt_spec *int_info, buffer_info *info, char *type_name, L4_ThreadId_t tid);

void collect_samples(int sampling_params, interrupt_spec *int_info, buffer_info *info, L4_ThreadId_t tid);//void (*fn)(void*));
void register_thread(L4_ThreadId_t tid_to_register);//void (*fn)(void*));
void new_input_link(L4_ThreadId_t my_thread, buffer_info *input_data);

int thread_index_raw(L4_Word_t input);
buffer_info* binfo_raw_thread(L4_Word_t id);
int buffer_info_index(buffer_info *info);

char *retrieve_data(buffer_info *info, L4_ThreadId_t my_id, int *no_record);
char *obtain_sample_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);
void copy_raw_record(char *sample_record, buffer_info *dest_thread_buffer);
L4_Word_t insert_aggregation(buffer_info *info, char *record);
void copy_core_samples_over(buffer_info *info);
void copy_sample_window(char *sample_record, buffer_info *dest_thread_buffer);
void copy_sample_record(char *sample_record, buffer_info *dest_thread_buffer);

void get_write_index(buffer_info *info, L4_Word_t *address);
void change_write_index(buffer_info *info, L4_Word_t amount);
void write_index_operation(buffer_info *info, char operation, L4_Word_t *addr, L4_Word_t amount);
L4_Word_t get_reset_state(buffer_info *info, L4_Word_t index);
L4_Word_t all_threads_resetted(buffer_info *info);

int duration_factor(L4_Word_t rate_unit, L4_Word_t duration_unit);

void finish_stream(L4_ThreadId_t tid);

void suspend(ProfVector<L4_ThreadId_t> *vec);
void resume(ProfVector<L4_ThreadId_t> *vec, L4_ThreadId_t phase_id);
void kill(ProfVector<L4_ThreadId_t> *vec);

void profiling_init();
