#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include "sampling1.h"
#include <l4/arch/profpolicy.h>
#include <stdarg.h>

#include <mutex/mutex.h>

#include <l4/thread.h>
#include <l4/ipc.h>

volatile L4_Word_t matrix_index = 0;
volatile int thread_matrix[THREAD_QUANTITY][THREAD_QUANTITY];

volatile buffer_info *thread_buffers[THREAD_QUANTITY];
L4_Word_t thread_list[THREAD_QUANTITY];
L4_ThreadId_t list_of_reg_threads[THREAD_QUANTITY];

uintptr_t thread_ptrs[THREAD_QUANTITY];

volatile L4_ThreadId_t thread_tunit[6];
volatile L4_Word_t thread_list_index = 0;

volatile okl4_mutex_t t_mutex;

/*
	Initialize the matrix elements such that we will know
	if a buffer has been assigned to or not.
*/
void init_matrix()
{
	for(int i = 0;i < THREAD_QUANTITY;i++)
	{
		for(int j = 0;j < THREAD_QUANTITY;j++)
			thread_matrix[i][j] = 0;
	}
	
	for(int i = 0;i < 6;i++)
		thread_tunit[i] = L4_nilthread;
}

/* 	Keep a record of collection threads that are using certain interrupt
	mechanisms */
void assign_tid_tunit(L4_ThreadId_t coll_thread, L4_Word_t unit_assign)
{
	thread_tunit[unit_assign] = coll_thread;
}

L4_ThreadId_t tunit_tid(L4_Word_t unit_assign)
{
	return thread_tunit[unit_assign];
}

/* For a time unit, get the relevant thread index */
L4_Word_t tunit_index(L4_Word_t tunit)
{
	return thread_index(thread_tunit[tunit]);
}

/* 	
	As part of the thread registration, the first step is to
	enter a reference to that thread in the thread entry array
	
	Parameters:
		- sampler: the ID of the thread to register
		- buffer_addr: the address of the thread's sampling buffer
*/
L4_Word_t enter_thread_entry(L4_Word_t tid_register, uintptr_t fn_ptr, buffer_info *buffer)
{	
	//L4_Word_t tindex = thread_index_raw(tid_register);
	int tindex = thread_ptr_index(fn_ptr);
	if(tindex != -1)
	{
		thread_buffers[tindex] = buffer;
		return tindex;
	}
	else if(matrix_index <= THREAD_QUANTITY)
	{
		thread_buffers[matrix_index] = buffer;
		thread_ptrs[matrix_index] = fn_ptr;
		matrix_index++;
		//thread_list[matrix_index++] = tid_register;
		return matrix_index - 1;
	}
	
	printf("We have reached the maximum no. of threads\n");
	return -1;
}

L4_Word_t index_to_id(L4_Word_t index)
{
	if(index >= matrix_index)
		return -1;
	
	return thread_list[index];
}

L4_ThreadId_t thread_identifier(int index)
{
	if (index >= matrix_index)
		return L4_nilthread;
	
	return list_of_reg_threads[index];
}

/* Get the index of the thread ID passed */
L4_Word_t thread_index(L4_ThreadId_t input)
{
	for(int i = 0;i < matrix_index;i++)
	{
		if(list_of_reg_threads[i].raw == input.raw)
		{
			return i;
		}
	}
	
	return -1;
}

int thread_index_fn(void (*fn)(void*))
{
	int i;
	for(i = 0;i < matrix_index;i++)
	{

		if(thread_ptrs[i] == (uintptr_t)fn)
		{
			return i;
		}
	}
	
	return -1;
}

L4_Word_t thread_ptr_index(uintptr_t thread_ptr)
{
	for(int i = 0;i < matrix_index;i++)
	{
		if(thread_ptrs[i] == thread_ptr)
			return i;
	}
	
	return -1;
}

L4_Word_t thread_index_raw(L4_Word_t input)
{
	for(int i = 0;i < matrix_index;i++)
	{
		if(thread_list[i] == input)
			return i;
	}
	
	return -1;
}

#if 0
void create_prof_thread(void (*fn)(void*), void *args)
{	
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	
	L4_MsgAppendWord(&msg, (uintptr_t)fn);
	L4_MsgAppendWord(&msg, (uintptr_t)args);
	L4_MsgLoad(&msg);

	tag = L4_Send(thread_id_reg);
	printf("PROF_THREAD: I SHOULD APPEAR TWICE\n");
}

void add_to_thread_list(void *args)
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_ThreadId_t sender;
	
	uintptr_t fn_ptr;
	void *thread_args;
	
	while(1)
	{
		/* 
			message structure:
				- buffer address
		*/
		printf("in add thread list function\n");
		L4_MsgClear(&msg);
		tag = L4_Wait(&sender);
		L4_MsgStore(tag, &msg);

		fn_ptr = L4_MsgWord(&msg, 0);
		thread_args = (void*)L4_MsgWord(&msg, 1);
		
		L4_ThreadId_t new_tid;
		thread_ref_t new_thread = thread_create_priority(130, &new_tid);
	
		list_of_reg_threads[thread_list_index] = thread_l4tid(new_thread);
		thread_list_index++;
	
		thread_start(new_thread, (uintptr_t)fn_ptr, (uintptr_t)thread_args);

		//L4_MsgClear(&msg);		
		//L4_MsgAppendWord(&msg, index);
		//L4_MsgLoad(&msg);
		//L4_Reply(sender);

	}
		
}
#endif
/* Get the buffer info structure based on the raw ID passed */
buffer_info* binfo_raw_thread(L4_Word_t id)
{
	for(int i = 0;i < matrix_index;i++)
	{//Will need to inspect matrix and check whether the thread
	 //is allowed to collect information about the buffer.
		if(thread_list[i] == id)
			return (buffer_info*)thread_buffers[i];
	}
	
	return NULL;
}

buffer_info* binfo_fn(uintptr_t fn)
{
	for(int i = 0;i < matrix_index;i++)
	{
		if(thread_ptrs[i] == fn)
			return thread_buffer_info(i);
	}
	
	return NULL;
}

buffer_info* thread_buffer_info(L4_Word_t index)
{
	if(index < matrix_index)
	{
		return (buffer_info*)thread_buffers[index];
	}
	
	return NULL;
}

/*
	Store a reference to the location of any buffers being inspected by the given thread.
	The starting address of the buffer to inspect is effectively returned via the pointer.
	
	Parameters:
		- source_thread_index: the index of the thread wanting to link to a thread
		- destination_thread_index: the index of the thread being linked to.  This will usually
							 	 	correspond to a thread that produces output data.  A thread
							  		that doesn't produce output data implies that the thread
									will not have any links going to it as it won't have a
								 	thread_buffer_info structure associated with it.
*/

/*	May have to change it so that it is implied that the thread that called it wants to
	create the link to another thread */
void enter_matrix_entry(L4_Word_t source_thread_index, L4_Word_t destination_thread_index)
{

	/* 	Ensure that the thread we want to connect to actually
		has its own buffer info structure */
	if(thread_buffer_info(destination_thread_index) == NULL)
	{
		printf("destination_thread_index: Invalid thread index specified\n");
		return;
	}
	
	/*	Ensure the input thread is actually valid.  This is checked by
		seeing whether the index is less than the total number of sampling
		threads */
	if(source_thread_index >= matrix_index)
	{
		printf("source_thread_index: Invalid thread index specified\n");
		return;
	}

	//Here, we want to store a reference to the index of the buffer.
	//ATM, need to find a more reliable way of doing this though.
	//Also, the format of the matrix is: output, input ie. row, column.
	//That is, the rows correspond to the threads looking at a thread's output
	//while the columns indicate the inputs that the thread is currently using.
	thread_matrix[destination_thread_index][source_thread_index] = 1;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);

	thread_buffer->threads[source_thread_index] = thread_buffer->buffer;
	thread_buffer->connected_threads++;
}

void reestablish_links(L4_ThreadId_t tid_to_fix)
{
	buffer_info *info;
	L4_Word_t tindex_for_fixing = thread_index(tid_to_fix);
	for(int i = 0;i < matrix_index;i++)
	{
		if(thread_matrix[i][tindex_for_fixing] == 1)
		{
			info = thread_buffer_info(i);
			info->threads[tindex_for_fixing] = info->buffer;
		}
	}
}

void register_thread(L4_ThreadId_t tid_to_register, void (*fn)(void*))
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);

	// 0 means that the sampling thread will not produce any output.  It
	// will however use data from other threads
	L4_MsgAppendWord(&msg, tid_to_register.raw);
	L4_MsgAppendWord(&msg, (uintptr_t)fn);
	L4_MsgAppendWord(&msg, 0);
	L4_MsgLoad(&msg);
	
	tag = L4_Send(register_id);
}

L4_Word_t determine_size(buffer_info *info)
{//this should only apply to a collection spec
	//use this to allocate memory for the core window
	
	int sample_size = info->actual_chunk_size;

	/*
		If the time units for the sample collection and consolidation are the same,
		the amount of memory to allocate is always going to be:
		
		window length * sample size * number of windows
	*/
	if((info->interrupt_no == MSEC_TIMER && info->window_unit == UNIT_MSEC)
		|| (info->interrupt_no == SEC_TIMER && info->window_unit == UNIT_SEC)
		|| (info->interrupt_no == MIN_TIMER && info->window_unit == UNIT_MIN)
		|| (info->interrupt_no == HR_TIMER && info->window_unit == UNIT_HR))
	{
		return sample_size * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == MSEC_TIMER)
	{
		if(info->window_unit == UNIT_SEC)
			return sample_size * 513 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_MIN)
			return sample_size * 30780 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 1846800 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == SEC_TIMER)
	{
		if(info->window_unit == UNIT_MSEC)
			return -1; 
		else if(info->window_unit == UNIT_MIN)
			return sample_size * 60 * info->time_window * info->no_windows;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 3600 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == MIN_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC)
			return -1;
		else if(info->window_unit == UNIT_HR)
			return sample_size * 60 * info->time_window * info->no_windows;
	}
	else if(info->interrupt_no == HR_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC || info->window_unit == UNIT_HR)
			return -1;
	}
	
	return -1;
}

void reg_thread(void *args)
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_ThreadId_t sender;
	
	L4_Word_t tid_register;
	uintptr_t fn_ptr;
	L4_Word_t buffer;
	L4_Word_t index;
	while(1)
	{
		/* 
			message structure:
				- buffer address
		*/
		L4_MsgClear(&msg);
		tag = L4_Wait(&sender);
		L4_MsgStore(tag, &msg);
		tid_register = L4_MsgWord(&msg, 0);
		fn_ptr = L4_MsgWord(&msg, 1);
		buffer = L4_MsgWord(&msg, 2);
		
		if(buffer == 0)
		{
			index = enter_thread_entry(tid_register, (uintptr_t)fn_ptr, NULL); 
		}
		else
		{
			index = enter_thread_entry(tid_register, (uintptr_t)fn_ptr, (buffer_info*)buffer);
		}
		
		L4_MsgClear(&msg);		
		L4_MsgAppendWord(&msg, index);
		L4_MsgLoad(&msg);
		L4_Reply(sender);

	}
}

/*
 *	Arguments:
 *		- input_ptr: Used to store a pointer to a thread's output buffer
 *		- output_data: The thread's buffer to point to
 *
 */
//void new_input_link(L4_Word_t input_source_index, L4_Word_t input_getter_index)
//void new_input_link(L4_ThreadId_t thread_linking, void (*fn)(void*))
void new_input_link(void (*fn1)(void*), void (*fn2)(void*))
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	
	
	//L4_MsgAppendWord(&msg, thread_index(thread_linking));
	//L4_MsgAppendWord(&msg, input_getter_index);
	L4_MsgAppendWord(&msg, thread_ptr_index((uintptr_t)fn1));
	L4_MsgAppendWord(&msg, thread_ptr_index((uintptr_t)fn2));
	L4_MsgLoad(&msg);

	tag = L4_Send(matrix_id);
}

void enter_link(void *args)
{
	L4_MsgTag_t tag;
	L4_Msg_t msg;	
	L4_ThreadId_t sender;
	
	L4_Word_t input_source_index = 0;
	L4_Word_t input_getter_index = 0;

	while(1)
	{
		L4_MsgClear(&msg);

		tag = L4_Wait(&sender);
		L4_MsgStore(tag, &msg);

		input_source_index = L4_MsgWord(&msg, 0);
		input_getter_index = L4_MsgWord(&msg, 1);
		L4_MsgClear(&msg);

		enter_matrix_entry(input_source_index, input_getter_index);

		buffer_info *buf1 = thread_buffer_info(input_getter_index);
		buffer_info *buf2 = thread_buffer_info(input_source_index);
			
		//if the thread that we're connecting to is not consolidating and
		//we've specified that the buffer is to consolidate data, set the state
		//to non-consolidation given that we can't consolidate data in the same
		//way as consolidation while collecting samples.
		
		//This is the likely source of the bug
		if(buf2 != NULL && buf2->storage_mode == CONSOLIDATE)
		{
			if(buf1 != NULL && buf1->storage_mode == NO_CONSOLIDATE)
			{
				buf2->storage_mode == NO_CONSOLIDATE;
			}
		}
		
		if(buf1 != NULL && buf1->storage_mode == CONSOLIDATE)
		{
			L4_MsgAppendWord(&msg, buf1->core_size); //size of core buffer
			L4_MsgAppendWord(&msg, buf1->size);		 //sampling buffer size
			L4_MsgAppendWord(&msg, 1);				 //consolidating or not
			L4_MsgAppendWord(&msg, buf1->time_window);
			L4_MsgAppendWord(&msg, buf1->no_windows);
			L4_MsgLoad(&msg);
		}
		else if(buf1 != NULL && buf1->storage_mode == NO_CONSOLIDATE)
		{//within this, add another condition to return other information if the current
		 //buffer is aggregating samples
			L4_MsgAppendWord(&msg, 40);
			L4_MsgAppendWord(&msg, 0);
			L4_MsgAppendWord(&msg, 0);
			L4_MsgAppendWord(&msg, buf1->time_window);
			L4_MsgAppendWord(&msg, buf1->no_windows);
			L4_MsgLoad(&msg);
		}
		L4_Reply(sender);
	}
}

L4_Word_t sample_equals(buffer_info *info, char *sample1, char *sample2)
{
	/* 
		Sample equality is defined as the following:
			- same interrupt value
			- same information collected
			- same TID and/or inst, depending on what is collected
	*/
	
	//check each element -> only check either TID, instruction addr or both
	
	//if(L4interrupt_type(sample1) == L4interrupt_type(sample2) &&
	//	L4info_collect(sample1) == L4info_collect(sample2) &&
	//	L4sample_size(sample1) == L4sample_size(sample2))
	if(info->info_collect & INFO_TID && info->info_collect & INFO_INST)
	{//if so, we are collecting both TID and instruction addr info
		int first_1 = *((int*)sample1);
		int first_2 = *((int*)sample2);
		unsigned int second_1 = *((int*)(sample1 + 4));
		unsigned int second_2 = *((int*)(sample2 + 4));

		if(first_1 == first_2 && second_1 == second_2)
		{
			return 1;
		}
	}
	else if(info->info_collect & INFO_TID || info->info_collect & INFO_INST)
	{
		int first_1 = *((int*)sample1);
		int first_2 = *((int*)sample2);

		//printf("TID 1: %d\tTID 2: %d\n", first_1, first_2);

		if(first_1 == first_2)
			return 1;
	}

	return 0;
	
}

uintptr_t minimum_index_bounds(buffer_info *buffer)
{
	L4_Word_t non_null_pointers = 0;
	L4_Word_t null_char_pointers = 0;
	L4_Word_t scanned = 0;
	
	L4_Word_t min = ULONG_MAX;
	
	/*
		To obtain the minimum index, the value has to lie in between the write index and the end
		of the buffer.  In the case where all related indices are not within the bounds, this
		would indicate that we are in the starting copying state (ie. where samples can be written
		normally).
	*/
	
	for(L4_Word_t i = 0;i < matrix_index;i++)
	{
		if(buffer->moved[i])
		{
			//printf("we have a moving record\n");
			scanned = 1;
			non_null_pointers++;
			
			if(buffer->t_offsets[i] < min)
				//&& buffer->buffer + buffer->t_offsets[i] > (uintptr_t)buffer->write_index)
			{
				min = buffer->t_offsets[i];
			}
		}
	}
	
	/* Returned if there are no thread links at all */
	if(null_char_pointers == matrix_index)
		return NO_THREAD_LINKS;
	
	/* Returned if there are thread links and the threads in question have
		not started reading data yet */
	if(non_null_pointers == buffer->connected_threads && scanned == 0)
		return NO_BUFFER_SCANS;
	
	return min;
}

void get_write_index(buffer_info *info, L4_Word_t *address)
{
	write_index_operation(info, READ, address, 0);
}

void change_write_index(buffer_info *info, L4_Word_t amount)
{
	write_index_operation(info, WRITE, 0, amount);
}

void write_index_operation(buffer_info *info, char operation, L4_Word_t *addr, L4_Word_t amount)
{
	if(operation == READ)
		*addr = (uintptr_t)info->write_index;
	else
		info->write_index += amount;
		
}

void set_reset_state(buffer_info *info)
{
	for(L4_Word_t i = 0;i < matrix_index;i++)
	{
		if(info->threads[i] != NULL)
		{
			info->resetted[i] = 1;
		}
	}
}

L4_Word_t get_reset_state(buffer_info *info, L4_Word_t index)
{
	if(info->threads[index] != NULL)
	{
		return info->resetted[index];
	}
	
	return -1;
}

L4_Word_t all_threads_resetted(buffer_info *info)
{
	for(L4_Word_t i = 0;i < matrix_index;i++)
	{
		if(info->threads[i] != NULL && info->resetted[i] == 1)
		{
			return 0;
		}
	}
	
	return 1;
}

char *obtain_raw_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
							int *no_record)
{
	*no_record = 0;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	L4_Word_t current_offset = thread_buffer->t_offsets[source_thread_index];

	char *current_index = (char*)(thread_buffer->buffer + current_offset);
	L4_Word_t current_index_addr = (uintptr_t)current_index;
	L4_Word_t wrt_index;
	get_write_index(thread_buffer, &wrt_index);
	L4_Word_t thread_reset_state = get_reset_state(thread_buffer, source_thread_index);
	char *new_buffer = NULL;

	if(thread_reset_state == 0)
	{
		if(current_index == NULL)
		{
			L4_Stop_Thread(L4_Myself());
		}
		else if(current_index_addr >= wrt_index)
		{
			*no_record = 1;
		}
		//else if(*current_index == MSEC_END || *current_index == SEC_END ||
		//		*current_index == MIN_END || *current_index == HR_END ||
		//		*current_index == PROF_END)
		else if(memcmp(current_index, "T_END", 5) == 0 || memcmp(current_index, "S_END", 5) == 0 ||
				memcmp(current_index, "M_END", 5) == 0 || memcmp(current_index, "H_END", 5) == 0 ||
				memcmp(current_index, "END", 4) == 0)
		{//It must either be one of the end particular sampling tags (eg. SEC_END) or 
		 //a profiling tag.  A PROF_END tag should imply that collection has been finalized for 
		 //the given thread anyway.
			new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 6;
			*no_record = 1;
			return new_buffer;
		}
		else if(memcmp(current_index, "P1_END", 6) == 0 || memcmp(current_index, "P2_END", 6) == 0)
		{
			new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 7;
			*no_record = 1;
			return new_buffer;
		}
		else
		{
			//here, check that we can get the sample record.  If not, we switch to any thread.  Otherwise, obtain
			//data as normal.
			if(current_index_addr + thread_buffer->actual_chunk_size >= wrt_index)
			{
				*no_record = 1;
			}
			else
			{
				new_buffer = (char*)malloc(thread_buffer->actual_chunk_size);
				memcpy(new_buffer, (void*)current_index, thread_buffer->actual_chunk_size);
				thread_buffer->t_offsets[source_thread_index] += thread_buffer->actual_chunk_size;//L4sample_size(current_index);
				thread_buffer->moved[source_thread_index] = 1;
				
				return new_buffer;
			}
		}
	}
	
	if(thread_reset_state == 1)
	{
		if(*current_index == BUFFER_RESET)
		{//Upon encountering this tag, the read index is resetted.  If all threads have resetted,
		 //the buffer state is set to non-reset ie. samples are copied to the buffer as per normal.
			thread_buffer->t_offsets[source_thread_index] = 0;
			thread_buffer->no_scanned++;
			thread_buffer->moved[source_thread_index] = 0;
			thread_buffer->resetted[source_thread_index] = 0;
			//If all threads have scanned through the buffer, sample writing can 
			//take place as normal.
			if(all_threads_resetted(thread_buffer) == 1)
			{
				thread_buffer->no_scanned = 0;
				thread_buffer->reset = 0;
			}
			*no_record = 1;
		}
		else
		{
			new_buffer = (char*)malloc(thread_buffer->actual_chunk_size);
			memcpy(new_buffer, (void*)current_index, thread_buffer->actual_chunk_size);
			thread_buffer->t_offsets[source_thread_index] += thread_buffer->actual_chunk_size;
			thread_buffer->moved[source_thread_index] = 1;
			return new_buffer;
		}
	}
	
	new_buffer = (char*)malloc(1);
	*new_buffer = 'z';
	return new_buffer;
}

char *obtain_sample_window(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
							int *no_record)
{
	*no_record = 0;
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	L4_Word_t current_offset = thread_buffer->t_offsets[source_thread_index];
	char *current_index = (char*)(thread_buffer->buffer + current_offset);
	L4_Word_t current_index_addr = (uintptr_t)current_index;

	L4_Word_t wrt_index;
	get_write_index(thread_buffer, &wrt_index);
	L4_Word_t thread_reset_state = get_reset_state(thread_buffer, source_thread_index);
	
	L4_Word_t window_size = determine_size(thread_buffer) / thread_buffer->no_windows + 1;
	
	char *new_buffer;
	
	if(thread_reset_state == 0)
	{
		if(current_index == NULL)
		{
			printf("obtain_sample_window: This thread is not permitted to access the specified buffer\n");
			L4_Stop_Thread(L4_Myself());
		}
		else if(current_index_addr >= wrt_index)
		{
			*no_record = 1;
		}
		/*else if(*current_index == MSEC_END || *current_index == SEC_END ||
				*current_index == MIN_END || *current_index == HR_END ||
				*current_index == PROF_END)
		{
			new_buffer = "END";
			thread_buffer->t_offsets[source_thread_index]++;
			*no_record = 1;
			return new_buffer;
		}*/
		else if(memcmp(current_index, "T_END", 5) == 0 || memcmp(current_index, "S_END", 5) == 0 ||
				memcmp(current_index, "M_END", 5) == 0 || memcmp(current_index, "H_END", 5) == 0)
		{//It must either be one of the end particular sampling tags (eg. SEC_END) or 
		 //a profiling tag.  A PROF_END tag should imply that collection has been finalized for 
		 //the given thread anyway.
			new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 6;
			*no_record = 1;
			return new_buffer;
		}
		else if(memcmp(current_index, "P1_END", 6) == 0 || memcmp(current_index, "P2_END", 6) == 0)
		{
			new_buffer = "END";
			//thread_buffer->t_offsets[source_thread_index]++;
			thread_buffer->t_offsets[source_thread_index] += 7;
			*no_record = 1;
			return new_buffer;
		}
		else
		{
			if(current_index_addr + window_size > wrt_index)
			{
				*no_record = 1;
			}
			else
			{
				new_buffer = (char*)malloc(window_size);
				memcpy((void*)new_buffer, (void*)current_index, window_size);
				thread_buffer->t_offsets[source_thread_index] += window_size;
				thread_buffer->moved[source_thread_index] = 1;
				return new_buffer;
			}
		}
	}
	
	if(thread_reset_state == 1)
	{
		if(*current_index == BUFFER_RESET)
		{
			thread_buffer->t_offsets[source_thread_index] = 0;
			thread_buffer->no_scanned++;
			thread_buffer->moved[source_thread_index] = 0;
			
			thread_buffer->resetted[source_thread_index] = 0;
			
			if(all_threads_resetted(thread_buffer) == 1)
			{
				thread_buffer->no_scanned = 0;
				thread_buffer->reset = 0;
			}
			
			*no_record = 1;
			//current_index = (char*)(thread_buffer->buffer);
		}
		else if(*current_index == MSEC_END || *current_index == SEC_END ||
				*current_index == MIN_END || *current_index == HR_END ||
				*current_index == PROF_END)
		{
			new_buffer = "END";
			thread_buffer->t_offsets[source_thread_index]++;
			*no_record = 1;
			return new_buffer;
		}
		else
		{
			new_buffer = (char*)malloc(window_size);
			memcpy((void*)new_buffer, (void*)current_index, window_size);
			thread_buffer->t_offsets[source_thread_index] += window_size;
			thread_buffer->moved[source_thread_index] = 1;
			return new_buffer;
		}
	}
	
	new_buffer = (char*)malloc(1);
	*new_buffer = 'z';
	return new_buffer;
}

char *obtain_sample_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
							int *no_record)
{
	buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	
	if(thread_buffer->storage_mode == NO_CONSOLIDATE)
		return obtain_raw_record(destination_thread_index, source_thread_index, no_record);

	return obtain_sample_window(destination_thread_index, source_thread_index, no_record);

}

//params for new function: input thread, current thread, no_record
//char *retrieve_data(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, 
//							L4_Word_t *no_record)
//char *retrieve_data(void (*fn)(void*), L4_ThreadId_t current_thread, int *no_record)
char *retrieve_data(void (*fn1)(void*), void (*fn2)(void*), int *no_record)
{
	//buffer_info *thread_buffer = thread_buffer_info(destination_thread_index);
	L4_Word_t input_thread_index = thread_ptr_index((uintptr_t)fn1);
	L4_Word_t source_thread_index = thread_ptr_index((uintptr_t)fn2);//thread_index(current_thread);
	
	//return obtain_sample_record(destination_thread_index, source_thread_index, no_record);
	return obtain_sample_record(input_thread_index, source_thread_index, no_record);
}

/*
	Parameters:
		- buffer: the consolidated samples the buffer will be copied to
		- current_ptr: used as the "iterator" over the buffer (defined above)
		- record: the sample about to be consolidated
		- table: Meant to be used as means for storing consolidated data.  Probably
					not required
		- agg_code: Char code corresponding to the aggregation to perform over samples
*/
L4_Word_t insert_aggregation(buffer_info *info, char *record)//, char *buffer, char *current_ptr, char *record, char agg_code)
{
	/*
		For the given record, scan through the consolidation buffer. From the current scan pointer,
		scan through the buffer.  If we get beyond the write pointer of the consolidated buffer,
		add the sample to the buffer along with the corresponding value.  Otherwise, increment
		the value for the given sample.
	*/
	
	char *entry = (char*)info->intermediate_buffer;
	while((uintptr_t)entry < (uintptr_t)info->intermediate_ptr)
	{//incrementing current_ptr???? NO -> should only happen if the sample has not been found
		//entry -> intermediate buffer. Hence, need to do entry +4
		if(sample_equals(info, record, (char*)(entry + 4)) == 1)
		{
			//increment the count value appropriate (COUNT is the only operation
			//that is possible for consolidation of samples)
			if(info->info_collect & INFO_TID && info->info_collect & INFO_INST)
			{
				*((int*)(entry + 12)) += 1;
			}
			else if(info->info_collect & INFO_TID || info->info_collect & INFO_INST)
			{
				*((int*)(entry + 8)) += 1;
			}
			
			return 1;
		}
		
		entry += info->actual_chunk_size;
	}

	size_t sample_size = info->actual_chunk_size - 8;
	*((int*)entry) = info->type_identifier;
	entry += sizeof(int);
	memcpy((void*)entry, (void*)record, sample_size);
	entry += sample_size;
	*((int*)entry) = 1;
	info->intermediate_ptr += info->actual_chunk_size;
	
	return 0;
}

/************************************************************************************
*	Additional functions for the alternative implementation below					*
*																					*
*************************************************************************************/

void min_time_unit(buffer_info *info, int *curr_unit, L4_Word_t tindex)
{
	buffer_info *next_buffer;
	int iterate = 0;
	int found = 0;
	int index = tindex;
	printf("index of thread is %d\n", index);
	if(index == -1)
	{
		printf("This is an invalid sampling thread\n");
		L4_Stop_Thread(L4_Myself());
	}
	
	while(iterate < matrix_index)
	{
		if(thread_matrix[iterate][index] == 1)
		{
			printf("Current window unit is %d\n", info->window_unit);
			printf("Current window unit is %d\n", *curr_unit);
			found = 1;
			next_buffer = thread_buffer_info(iterate);
			assert(next_buffer != NULL);
			printf("next buffer unit is %d\n", next_buffer->window_unit);
			printf("actual time unit is %d\n", next_buffer->interrupt_no);
			if(next_buffer->storage_mode == CONSOLIDATE)
			{
				if(next_buffer->window_unit < info->window_unit)
				{
					*curr_unit = next_buffer->window_unit;
					printf("all tests: curr unit is %d\n", *curr_unit);
				}
			}
			
			//if the thread being connected to is not consolidating data at all,
			//the traversal simply continues.  Eventually, there will be "nodes"
			//that will have a time unit.  These correspond to the collection
			//specifications.
			min_time_unit(next_buffer, curr_unit, iterate);
		}
		
		iterate++;
	}
	
	if(found == 0)
	{	next_buffer = thread_buffer_info(index);
		printf("nothing1: curr_unit is %d\n", *curr_unit);
		if(next_buffer->interrupt_no < *curr_unit)
		{
			*curr_unit = info->interrupt_no;
			printf("nothing2: curr_unit is %d\n", *curr_unit);
		}
	}
}

void print_matrix()
{
	for(int i = 0; i < matrix_index;i++)
	{
		for(int j = 0;j < matrix_index;j++)
		{
			printf("%d\t", thread_matrix[i][j]);
		}
		
		printf("\n");
	}
}

//size: refers to the quantity of types to be used for the buffer
void initialize_buffer(interrupt_spec *int_info, buffer_info *info, char *type_name, void (*fn)(void*))
{
	if(int_info != NULL)
	{
		info->interrupt_no = int_info->interrupt_no;
		info->info_collect = int_info->info_collect;
		info->duration = int_info->duration;
		info->rate = int_info->rate;
		info->time_window =int_info->time_window;
		info->window_unit = int_info->window_unit;
		info->no_windows = int_info->no_windows;
		info->storage_mode = int_info->storage_mode;
		printf("storage mode of buffer is %d\n", info->storage_mode);
	
		switch(info->type_identifier)
		{
			case -4: //boolean
				info->actual_chunk_size = 1;
				break;
			case -3: //string
				info->actual_chunk_size = 50;
				break;
			case -2: //double
				info->actual_chunk_size = sizeof(double);
				break;
			case -1:
				info->actual_chunk_size = sizeof(int);
				break;
			case 0:
				info->actual_chunk_size = sizeof(int) * 2;
				break;
			case 1:
				info->actual_chunk_size = sizeof(int) * 3;
				break;
			case 2:
				info->actual_chunk_size = sizeof(int) * 4;
				break;
			default:
			{
				tinfo *info_about_type = get_stats(info->type_identifier);
				info->actual_chunk_size = memory_single_chunk((char*)info_about_type->tup_type_seq) + sizeof(int);//additional id field
			}
		}
	}
	
	int index_thread = thread_index_fn(fn);
	buffer_info *iter;
	int found = 0;
	
	for(int i = 0;i < matrix_index;i++)
	{//Allocation of memory for all inputs to this thread.  This checks to see whether
	 //there are inputs to the thread so that the memory can be allocated correctly.
		if(thread_matrix[index_thread][i] == 1)
		{
			found = 1;
			iter = thread_buffer_info(i);
			
			//if i'm consolidating, look at the windowing information.  Check compatibility
			//and allocate the appropriate amount of memory.  Before this, consolidation modes
			//must be compatible as well ie. if the thread being connected to is not consolidating,
			//logic would suggest that this thread shouldn't be in consolidation mode as there is no
			//notion of windowing unconsolidated data.
			if(iter->storage_mode == CONSOLIDATE)
			{
				if(info->window_unit == UNIT_SEC && info->interrupt_no == MSEC_TIMER)
				{
					if(info->no_windows < 7)
					{
						info->no_windows = 7;
						int_info->no_windows = 7;
					}
				}
			}
			else if(info->storage_mode == CONSOLIDATE)
			{
				//if so, we cannot consolidate data from a thread that is not
				//consolidating data, so change the mode appropriately
				info->storage_mode == NO_CONSOLIDATE;
			}

			info->remaining_size = info->actual_chunk_size;
			
			if(int_info != NULL)
			{
				//Logic would suggest that if the thread is consolidating, the thread being
				//connected to is also consolidating.  This will be checked upon making a thread
				//link
				if(info->storage_mode == CONSOLIDATE)
				{	
					//Get the required memory for the consolidation buffer.  Calculations based on
					//"worst case" scenario
					if(info->window_unit == UNIT_SEC && info->interrupt_no == MSEC_TIMER)
					{
						if(info->no_windows < 7)
						{
							info->no_windows = 7;
							int_info->no_windows = 7;
						}
					}
					
					int win_factor = window_factor(info);
					info->time_window *= win_factor;
					
					if(info->time_window < iter->time_window)
					{
						info->time_window = iter->time_window;
					}

					while(info->time_window % iter->time_window != 0)
					{
						info->time_window++;
					}

					L4_Word_t new_size = iter->core_size * 
										 (info->time_window / iter->time_window) *
										 info->no_windows;

					info->core_size = iter->core_size * 
										 (info->time_window / iter->time_window);
					info->buffer = (char*)realloc((void*)info->buffer, info->size + new_size + sizeof(char));
					info->size += new_size + sizeof(char);
					info->write_index = info->buffer;
				}
				else if(iter->storage_mode == CONSOLIDATE)
				{//If the thread we are connecting to is consolidating, we should at the very least allocate
				 //data for the sampling window.
				 
					info->buffer = (char*)realloc((void*)info->buffer, iter->size);
					info->size += iter->size;
					info->write_index = info->buffer;
				}
				else
				{
					/* allocate a fixed amount of memory -> 130KB thereabouts => half of one buffer */
					info->buffer = (char*)realloc((void*)info->buffer, info->size + 131048);
					info->size += 131048;
					info->write_index = info->buffer;
				}
			}
			else
			{
				printf("Interrupt information not supplied\n");
				L4_Stop_Thread(L4_Myself());
			}
		}
	}
	
	if(found == 0)
	{//This must be a collection specification, so allocate memory based on the interrupt
	 //information supplied.
		if(info->storage_mode == CONSOLIDATE)
		{
			//printf("About to create consolidated memory. SO I KNOW IT'S INITTING\n");
			L4_Word_t consol_size = determine_size(info);
			//printf("consolidation size: %ld\n", consol_size);
			
			/*
				memory for storing count data is already used in buffer calculations
			*/
			info->buffer = (char*)malloc(consol_size + int_info->no_windows + sizeof(char));
			info->write_index = info->buffer;
			info->size = consol_size + int_info->no_windows + sizeof(char);
						 
			/*
				core_samples -> time window, including additional field
				
				size of a single window:
				consol_size/number of windows + (number of possible records in a window * size of field)
			*/
			
			//what i want to do here is only collect the required information in the first buffer. The
			//intermediate buffer will then have the required memory capacity (including the additional
			//fields for type id and count
			info->core_samples = (char*)malloc(consol_size / int_info->no_windows - (window_constant(info) * 8) + sizeof(char));
			info->core_size = consol_size / int_info->no_windows - (window_constant(info) * 8) + sizeof(char);
			info->core_pointer = info->core_samples;
		}
		else
		{
			info->buffer = (char*)malloc(131055);
			info->size = 131055;
			info->write_index = info->buffer;
		}
	}
	
	L4_MsgTag_t tag;
	L4_Msg_t msg;
	L4_MsgClear(&msg);
	L4_MsgAppendWord(&msg, L4_Myself().raw);
	L4_MsgAppendWord(&msg, (uintptr_t)fn);
	L4_MsgAppendWord(&msg, (L4_Word_t)info);
	
	L4_MsgLoad(&msg);
	
	//A message is sent to the matrix thread for registration.

	tag = L4_Send(register_id);
	
}

int memory_single_chunk(char *sequence)
{
	int i;
	int size = 0;
	for(i = 0;i < strlen(sequence);i++)
	{
		switch(sequence[i])
		{//NEED TO ADD BOOLEAN CASE
			case INT:
				size += sizeof(int);
				break;
			case DOUBLE:
				size += sizeof(double);
				break;
			case BOOLEAN:
				size += 1;
				break;
			case STRING:
				size += 50;
				break;
		}
	}
	return size;
}

/*
	When connecting to a thread, i need to determine the most fine grained time unit
	that is being used.  This will allow me to allocate data based on a worst case scenario
	where only unique samples are being generated.  For instance, if the current thread
	is providing consolidation at the granularity of hours and the destination thread is
	collecting data at the timer tick granularity, the current thread needs to allocate
	the following amount of memory:
	
	window length * 513 * 60 * 60 * number of windows
	
	This will allow us to cover the worst case situation of only unique samples being generated
	due to the fact that we cannot guarantee that a sampling run will not produce unique
	samples.  Furthermore, i am assuming window lengths and numbers are compatible in relation 
	to being able to fit all samples generated by the destination thread in a single copy run.
	
	Of course, i should not be able to consolidate if the time unit of the current thread
	is of finer granularity than the destination thread.
*/
int window_factor(buffer_info *info)
{
	int tunit_analysis;
	if(info->storage_mode == CONSOLIDATE)
		tunit_analysis = info->window_unit;
	else
		tunit_analysis = info->interrupt_no;

	int unit = tunit_analysis;
	min_time_unit(info, &unit, thread_index(L4_Myself()));
	
	if((tunit_analysis == UNIT_MSEC && unit == UNIT_MSEC) ||
		(tunit_analysis == UNIT_SEC && unit == UNIT_SEC) ||
		(tunit_analysis == UNIT_MIN && unit == UNIT_MIN) ||
		(tunit_analysis == UNIT_HR && unit == UNIT_HR))
	{
		return 1;
	}
	else if(tunit_analysis == UNIT_MSEC)
	{		
		if(unit == UNIT_SEC || unit == UNIT_MIN || unit == UNIT_HR)
		{
			printf("UNIT_TICK: Cannot provide consolidation based on units\n");
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_SEC)
	{
		if(unit == UNIT_MSEC)
			return 513;
		else if(unit == UNIT_MIN || unit == UNIT_HR)
		{
			printf("UNIT_SEC: Cannot provide consolidation based on units\n");
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_MIN)
	{
		if(unit == UNIT_SEC)
			return 60;
		else if(unit == UNIT_MSEC)
			return 60 * 513;
		else if(unit == UNIT_HR)
		{
			printf("UNIT_MIN: Cannot consolidate based on units\n");	
			return -1;
		}
	}
	else if(tunit_analysis == UNIT_HR)
	{
		if(unit == UNIT_MSEC)
			return 513 * 60 * 60;
		else if(unit == UNIT_SEC)
			return 60 * 60;
		else if(unit == UNIT_MIN)
			return 60;
	}
	
	return -1;
}

void initialize_binfo(buffer_info *info)
{
	info->interrupt_no = 0;
	info->info_collect = 0;
	info->thread_index = 0;
	info->duration = 0;
	info->rate = 0;
	info->time_window = 0;
	info->window_unit = 0;
	info->no_windows = 0;

	info->storage_mode = NO_CONSOLIDATE;
	info->core_size = 0;

	for(int i = 0;i < THREAD_QUANTITY;i++)
	{
		info->threads[i] = NULL;
		info->moved[i] = 0;
		info->resetted[i] = 0;
		info->t_offsets[i] = 0;
	}
	
	info->size = 0;
	info->reset = 0;
	
	info->connected_threads = 0;
	info->no_scanned = 0;
	
	info->type_identifier = -5;
	
	info->dropped_samples = 0;
	info->actual_chunk_size = 0;
	info->remaining_size = info->chunk_size;
}

void collect_samples(int sampling_params, interrupt_spec *int_info, buffer_info *info, void (*fn)(void*))
{
	//See if the buffer has memory allocated.  If not, allocate the memory
	if(info->buffer == NULL)
	{
		L4_MsgTag_t tag;
		L4_Msg_t msg;	
		L4_MsgClear(&msg);
			
		initialize_binfo(info);
		
		if(int_info->info_collect & INFO_TID &&
			int_info->info_collect & INFO_INST)
		{
			if(int_info->storage_mode == CONSOLIDATE)
			{
				info->type_identifier = 2;
			}
			else
			{
				info->type_identifier = 1;
			}
		}
		else if(int_info->info_collect & INFO_TID ||
				int_info->info_collect & INFO_INST)
		{
			if(int_info->storage_mode == CONSOLIDATE)
			{
				info->type_identifier = 1;
			}
			else
			{
				info->type_identifier = 0;
			}
		}
		else
		{
			printf("INITIALIZE BUFFER: Collection parameters not specified properly\n");
			return;
		}
		
		//info->buffer_type = RAW;
		initialize_buffer(int_info, info, NULL, fn);
		
		tag = L4_Receive(register_id);
		L4_MsgStore(tag, &msg);
		
		L4_Word_t given_index = L4_MsgWord(&msg, 0);
		assert(given_index == thread_index_fn(fn));
	}
	
	L4_ThreadId_t id_of_reg = list_of_reg_threads[thread_index_fn(fn)];

	
	switch(int_info->interrupt_no)
	{
		case MSEC_TIMER:
			assign_tid_tunit(id_of_reg, TID_MSEC);
			break;
		case SEC_TIMER:
			assign_tid_tunit(id_of_reg, TID_SEC);
			break;
		case MIN_TIMER:
			assign_tid_tunit(id_of_reg, TID_MIN);
			break;
		case HR_TIMER:
			assign_tid_tunit(id_of_reg, TID_HR);
			break;
		case PMC1:
			assign_tid_tunit(id_of_reg, TID_PMC1);
			break;
		case PMC2:
			assign_tid_tunit(id_of_reg, TID_PMC2);
			break;
	}
	
	//Perform the system call to collect samples
	L4_ProfPolicy(sampling_params, int_info);
}

void print_intr_info(buffer_info *thread_buffer)
{
	printf("interrupt no:%d\n", thread_buffer->interrupt_no);
	printf("info collect: %d\n", thread_buffer->info_collect);
	printf("duration: %d\n", thread_buffer->duration);
	printf("rate: %d\n", thread_buffer->rate);
	printf("time_window: %d\n", thread_buffer->time_window);
	printf("window_unit: %d\n", thread_buffer->window_unit);
	printf("no_windows: %d\n", thread_buffer->no_windows);
	printf("storage_mode: %d\n", thread_buffer->storage_mode);
}

L4_Word_t window_constant(buffer_info *info)
{//this should only apply to a collection spec
	/*
		If the time units for the sample collection and consolidation are the same,
		the amount of memory to allocate is always going to be:
		
		window length * sample size * number of windows
	*/
	if((info->interrupt_no == MSEC_TIMER && info->window_unit == UNIT_MSEC)
		|| (info->interrupt_no == SEC_TIMER && info->window_unit == UNIT_SEC)
		|| (info->interrupt_no == MIN_TIMER && info->window_unit == UNIT_MIN)
		|| (info->interrupt_no == HR_TIMER && info->window_unit == UNIT_HR))
	{
		return 1;
	}
	else if(info->interrupt_no == MSEC_TIMER)
	{
		if(info->window_unit == UNIT_SEC)
			return 513;
		else if(info->window_unit == UNIT_MIN)
			return 30780;
		else if(info->window_unit == UNIT_HR)
			return 1846800;
	}
	else if(info->interrupt_no == SEC_TIMER)
	{
		if(info->window_unit == UNIT_MSEC)
			return -1; 
		else if(info->window_unit == UNIT_MIN)
			return 60;
		else if(info->window_unit == UNIT_HR)
			return 3600;
	}
	else if(info->interrupt_no == MIN_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC)
			return -1;
		else if(info->window_unit == UNIT_HR)
			return 60;
	}
	else if(info->interrupt_no == HR_TIMER)
	{
		if(info->window_unit == UNIT_MSEC || info->window_unit == UNIT_SEC || info->window_unit == UNIT_HR)
			return -1;
	}
	
	return -1;
}

char *get_record(char *buff_ptr, int size)
{
	char *new_record = (char*)malloc(size);
	memcpy(new_record, buff_ptr, size);
	return new_record;
}

//void finish_stream(L4_ThreadId_t tid)
void finish_stream(void (*fn)(void*))
{
	//buffer_info *info = binfo_raw_thread(tid.raw);
	buffer_info *info = binfo_fn((uintptr_t)fn);
	L4_Word_t min_index = minimum_index_bounds(info);
	int first_time = 0;
	
	if(info->reset == 0)
	{
		if(info->write_index + 4 >= info->buffer + info->size)
		{
			if(min_index != NO_THREAD_LINKS && min_index != NO_BUFFER_SCANS &&
				info->buffer + 4 < info->buffer + min_index)
			{
				first_time = 1;
			}

			set_reset_state(info);
			info->reset = 1;
		}
		else
		{
			printf("END OF RECORD WRITTEN\n");
			char *end = "END";
			memcpy((void*)info->write_index, end, strlen(end) + 1);
			info->write_index += 4;
		}
	}
	
	if(info->reset == 1)
	{
		if(first_time == 1)
		{
			first_time = 0;
			*(info->write_index) = BUFFER_RESET;
			info->write_index = info->buffer;
		}
		else if(info->write_index + 1 < info->buffer + min_index)
		{
			//*(info->write_index) = END_OF_RECORDS;
			memcpy((void*)info->write_index, "END", strlen("END") + 1);
			info->write_index += 4;
		}
		else
		{
			printf("about to drop a sample\n");
			info->dropped_samples++;
		}
	}
}

L4_ThreadId_t create_temp_id(L4_Word_t raw_id)
{
	L4_ThreadId_t new_id;
	new_id.raw = raw_id;
	
	return new_id;
}

void suspend(void (*fn)(void*))
{	
	//get buffer_info for thread, if = 0, it must be a processing thread
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);

	if(tbuffer == NULL)
	{
		L4_ThreadId_t thr_suspend;
		thr_suspend.raw = index_to_id(tid_search);
		L4_Stop_Thread(thr_suspend);
	}
	else if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_suspend;
		thr_suspend.raw = index_to_id(tid_search);
		L4_Stop_Thread(thr_suspend);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = index_to_id(tid_search);
		L4_ProfPolicy(SAMPLING_SUSPEND, &tid_spec);
	}
}

void resume(void (*fn)(void*))
{
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);
	
	if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_resume;
		thr_resume.raw = index_to_id(tid_search);
		L4_Start(thr_resume);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = index_to_id(tid_search);
		L4_ProfPolicy(SAMPLING_RESUME, &tid_spec);
	}
}

void kill(void (*fn)(void*))
{
	L4_Word_t tid_search = thread_ptr_index((uintptr_t)fn);
	buffer_info *tbuffer = thread_buffer_info(tid_search);
	
	if(tbuffer->interrupt_no == 0)
	{
		//it is not a probe, so use the default scheduler
		L4_ThreadId_t thr_kill;
		thr_kill.raw = index_to_id(tid_search);
		//L4_ThreadControl(thr_kill, L4_nilspace, NULL, NULL, NULL, NULL, 0);
	}
	else
	{
		interrupt_spec tid_spec;
		tid_spec.tid = index_to_id(tid_search);
		L4_ProfPolicy(SAMPLING_KILL, &tid_spec);
	}
}
