#ifndef _LIBTRACE_SAMPBUFFER_H_
#define _LIBTRACE_SAMPBUFFER_H_

#define SAMPLING_THREAD 11244
#define THREAD_QUANTITY 100

#define TID_MSEC	0
#define TID_SEC		1
#define TID_MIN		2
#define TID_HR		3
#define TID_PMC1	4
#define TID_PMC2	5

#define NO_THREAD_LINKS	0
#define NO_BUFFER_SCANS 1

#define NO_USER_GEN_RECORDS	0
#define AGG_SAMPLES -1
#define USER_RECORDS -2

#define RAW		'0'
#define AGGN	'1'
#define TUPLE	'2'

#define END_OF_PERIOD		'Y'
#define END_OF_RECORDS		'Z'
#define NO_CONSOLIDATE	0
#define CONSOLIDATE		1
#define WINDOW_END '0'

#define READ '0'
#define WRITE '1'

#define NO_RECORD_WRITES	0
#define NORMAL_WRITE		1
#define BUFFER_RESETTED		2

#define NO_TYPE	'\0'
#define INT		'I'
#define DOUBLE	'D'
#define STRING	'S'
#define BOOLEAN 'B'

typedef struct
{
	int volatile interrupt_no;
	int volatile info_collect;
	int volatile thread_index;
	int volatile duration;
	int volatile rate;
	int volatile time_window;
	int volatile window_unit;
	int volatile no_windows;
	int volatile storage_mode;

	char volatile buffer_type;
	
	/* Buffer to store raw samples for a fixed time length.  Only used where
		samples are being consolidated. */
	char volatile *core_samples;
	
	/* The current write index of the raw sampling buffer */
	char volatile *core_pointer;
	
	/* Size of the core buffer */
	L4_Word_t volatile core_size;
	
	/* Stores the relevant type information for the stream*/
	char volatile *type_information;
	
	/* Type information stored in stream. Either a basic type (in which information is  */
	tinfo volatile *stream_type;
	int volatile type_identifier;
	
	char volatile *intermediate_buffer;
	char volatile *intermediate_ptr;

	/* The current buffer writing index */ 
	char volatile *write_index;
	
	/* The sampling buffer (which gets malloc'ed).  In consolidation mode,
		this is used as the consolidated buffer.  Otherwise, this is used
		as the buffer for storing raw samples. */
	char volatile *buffer;
	
	/* An array to keep track of the threads inspecting the buffer */
 	char volatile *threads[THREAD_QUANTITY];
 	
 	/* Offsets of threads connected to the buffer */
 	L4_Word_t volatile t_offsets[THREAD_QUANTITY];
 	
 	/* A simple array that determines whether an inspecting thread has
 	   started scanning or not */
 	int volatile moved[THREAD_QUANTITY];
 	
 	/* The size of the sampling buffer */
	L4_Word_t size;
	
	/* State variable to determine whether the write index has gone back to
		the beginning of the buffer */
	L4_Word_t volatile reset;
	
	/* Maintain current reset states of all sampling threads */
	L4_Word_t volatile resetted[THREAD_QUANTITY];
		
	/* The number of threads linked to this particular buffer (besides the 
		thread that the buffer is linked to */
	L4_Word_t volatile connected_threads;
	
	/* For all the connected threads, how many of them have scanned through the buffer */
	L4_Word_t volatile no_scanned;
	
	/* Counter to count the number of samples dropped due to insufficient space */
	L4_Word_t volatile dropped_samples;
	
	/* For writing and reading from the buffer purely in user space, the chunk size
	   determines the size of the record to be obtained from the buffer */
	size_t volatile chunk_size;
	
	size_t volatile actual_chunk_size;
	
	size_t volatile remaining_size;
}buffer_info;

extern int counter;

#define FIRST_SAMPLE(buffer)	buffer + sizeof(char) * 2 + sizeof(L4_Word_t) * 6
#define SECOND_SAMPLE(buffer)	FIRST_SAMPLE(buffer) + sizeof(char) + sizeof(L4_Word_t)
#define THIRD_SAMPLE(buffer)	SECOND_SAMPLE(buffer) + sizeof(char) + sizeof(L4_Word_t)

//Obtaining individual elements: Use the reference to the sample itself, not
//the struct storing this information.  The thread's read index should be modified
//when the whole sample record is obtained beforehand.
L4_INLINE int L4samp_get_tid(char *buffer)
{
	//Make sure that there is actually a TID record.  The TID record does not
	//exist in the case where the scanning of the sample type immediately
	//results in the record for the current instruction address
	if(*(FIRST_SAMPLE(buffer)) == SAMPLE_TID)
	{
		return *((int*)(FIRST_SAMPLE(buffer) + sizeof(char)));
	}
	else
	{
		return -1;
	}
}

//Get the information relating to the current instruction that was executing
L4_INLINE L4_Word_t L4samp_get_inst(char *buffer)
{
	//Make sure there is actually an instruction record.  In this case,
	//increment up to the first tag.  If the tag is an instruction tag, then get the
	//record.
	if(*(FIRST_SAMPLE(buffer)) == INSTN_ADDR)
	{
		//get the information
		return *(FIRST_SAMPLE(buffer) + sizeof(char));
	}
	else if(*(SECOND_SAMPLE(buffer)) == INSTN_ADDR)
	{
		//get the information
		return *(SECOND_SAMPLE(buffer) + sizeof(char));
	}
	else
	{
		return -1;
	}
}

L4_INLINE L4_Word_t L4info_collect(char *buffer)
{
	return *((L4_Word_t*)(buffer + sizeof(char) + sizeof(L4_Word_t) * 4));
}

L4_INLINE L4_Word_t L4sample_size(char *buffer)
{
	return *((L4_Word_t*)(buffer + sizeof(char) + sizeof(L4_Word_t)));
}

L4_INLINE L4_Word_t L4interrupt_type(char *buffer)
{
	return *((L4_Word_t*)(buffer + sizeof(char) + sizeof(L4_Word_t) * 2));
}

L4_INLINE L4_Word_t L4samp_id(char *buffer)
{
	return *((L4_Word_t*)(buffer + sizeof(char) + sizeof(L4_Word_t) * 3));
}

L4_INLINE L4_Word_t L4tid_belong(char *record)
{
	return *((L4_Word_t*)(record + sizeof(char)));
}

L4_INLINE char *aggregation_info(char *record)
{
	return record + sizeof(char) + sizeof(L4_Word_t) * 5;
}

L4_INLINE char L4aggregation_type(char *record)
{
	return *aggregation_info(record);
}

L4_INLINE L4_Word_t L4agg_value(char *record)
{
	if(L4aggregation_type(record) != AGG_AVG)
	{
		return *((L4_Word_t*)(aggregation_info(record) + sizeof(char)));
	}
	
	L4_Word_t average = *((L4_Word_t*)(aggregation_info(record) + sizeof(char))) / 
					*((L4_Word_t*)(aggregation_info(record) + sizeof(char) + sizeof(L4_Word_t)));
	
	return average;
}

//initialization functions
void init_matrix(void);
void reestablish_links(L4_ThreadId_t tid_to_fix);

//registration functions
//void register_thread(L4_ThreadId_t tid_to_register);
void register_thread(L4_ThreadId_t tid_to_register, void (*fn) (void*));

void reg_thread(void *args);
void enter_link(void *args);
//void new_input_link(L4_Word_t input_source_index, L4_Word_t input_getter);
//void new_input_link(L4_ThreadId_t thread_linking, void (*fn)(void*));
void new_input_link(void (*fn1)(void*), void (*fn2)(void*));
void assign_tid_tunit(L4_ThreadId_t coll_thread, L4_Word_t unit_assign);
buffer_info* thread_buffer_info(L4_Word_t index);

//void add_to_thread_list(void *args);
//void create_prof_thread(void (*fn)(void*), void *args);

//thread information retrieval functions
L4_Word_t thread_index(L4_ThreadId_t sampler);
int thread_index_fn(void (*fn)(void*));
int fn_index(uintptr_t addr);

L4_Word_t thread_ptr_index(uintptr_t thread_ptr);

L4_Word_t thread_index_raw(L4_Word_t input);
L4_Word_t tunit_index(L4_Word_t tunit);
L4_ThreadId_t tunit_tid(L4_Word_t unit_assign);
L4_Word_t thread_indices(L4_ThreadId_t sampler);
void enter_matrix_entry(L4_Word_t output_thread_index, L4_Word_t input_thread_index);
//L4_Word_t enter_thread_entry(L4_Word_t tid_register, buffer_info *buffer);
L4_Word_t enter_thread_entry(L4_Word_t tid_register, uintptr_t fn_ptr, buffer_info *buffer);
L4_Word_t minimum_index_bounds(buffer_info *buffer);

void copy_sample_record(char *sample_record, buffer_info *dest_thread_buffer);
void copy_sample_window(char *sample_record, buffer_info *dest_thread_buffer);
void copy_raw_record(char *sample_record, buffer_info *dest_thread_buffer);

void copy_sample_elements(buffer_info *dest_buffer, char *sample_record, L4_Word_t *entry, L4_Word_t head);
void copy_core_samples_over(buffer_info *info);
buffer_info* binfo_raw_thread(L4_Word_t id);
buffer_info* binfo_fn(uintptr_t fn);
L4_Word_t index_to_id(L4_Word_t index);

//char *retrieve_data(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, L4_Word_t *no_record);
//char *retrieve_data(void (*fn)(void*), L4_ThreadId_t current_thread, int *no_record);
char *retrieve_data(void (*fn1)(void*), void (*fn2)(void*), int *no_record);
char *obtain_sample_window(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);
char *obtain_raw_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);
char *obtain_sample_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);
char *obtain_window(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);
L4_Word_t write_record(L4_ThreadId_t tid, void *data, size_t data_size);
void finish_stream(void (*fn)(void*));
//void finish_stream(L4_ThreadId_t tid);
char *get_sample_record(L4_Word_t destination_thread_index, L4_Word_t source_thread_index, int *no_record);

void write_index_operation(buffer_info *info, char operation, L4_Word_t *address, L4_Word_t amount);
void get_write_index(buffer_info *info, L4_Word_t *address);
void change_write_index(buffer_info *info, L4_Word_t amount);

L4_Word_t insert_aggregation(buffer_info *info, char *record);
L4_Word_t determine_size(buffer_info *info);
L4_Word_t sample_equals(buffer_info *info, char *sample1, char *sample2);

L4_Word_t all_threads_resetted(buffer_info *info);
void set_reset_state(buffer_info *info);
L4_Word_t get_reset_state(buffer_info *info, L4_Word_t index);

/* 	Current index of matrix entry.  Will also tell us the number of 
	sampling threads being used */
extern volatile L4_Word_t matrix_index;

/* Thread index matrix */
extern volatile buffer_info *thread_buffers[THREAD_QUANTITY];

/* Thread indices */
extern L4_Word_t thread_list[THREAD_QUANTITY];
extern L4_ThreadId_t list_of_reg_threads[THREAD_QUANTITY];
extern uintptr_t thread_ptrs[THREAD_QUANTITY];

/* Connectivity matrix */
extern volatile int thread_matrix[THREAD_QUANTITY][THREAD_QUANTITY];

/* Matrix for associating collection thread with time unit. Only 6 collection
	threads/mechanisms simultaneously. */
extern volatile L4_ThreadId_t thread_tunit[6];

L4_ThreadId_t thread_identifier(int index);

/* Used to store the ids of the registration and matrix threads */
extern L4_ThreadId_t register_id;
extern L4_ThreadId_t matrix_id;
extern L4_ThreadId_t interrupt_id;
extern L4_ThreadId_t thread_id_reg;

/* State variable to signify the record currently being obtained */
extern char current_sample_record;
extern L4_Word_t remaining_size;
extern L4_Word_t current_tid;

extern L4_ThreadId_t samp_server_id;
extern volatile L4_Word_t thread_list_index;
extern volatile okl4_mutex_t t_mutex;

//Get information about the current call stack.  The relevant addresses
//are stored in elements ie. the developer has to supply a buffer.
L4_INLINE void L4samp_get_callstack(char *buffer, L4_Word_t *elements)
{
	//the information for the callstack will always be the final
	//record in a sample, meaning that if the callstack is the first
	//record (ie. in FIRST_SAMPLE(buffer)), it is the only information
	//being collected for that sample.
	char *pointer;
	if(*(FIRST_SAMPLE(buffer)) == CALLSTACK_BEGIN)
	{
		pointer = FIRST_SAMPLE(buffer) + sizeof(char);
	}
	else if(*(SECOND_SAMPLE(buffer)) == CALLSTACK_BEGIN)
	{
		pointer = SECOND_SAMPLE(buffer) + sizeof(char);
	}
	else if(*(THIRD_SAMPLE(buffer)) == CALLSTACK_BEGIN)
	{
		pointer = THIRD_SAMPLE(buffer) + sizeof(char);
	}
	else
	{
		printf("There is no callstack record\n");
		return;
	}
	
	L4_Word_t current_element;
	
	do
	{
		current_element = *((L4_Word_t*)pointer);
		*elements = current_element;
		elements++;
		pointer += sizeof(L4_Word_t);
		
	}while(current_element != 0);
	
	//Signify the end of the callstack with the value 0
	*elements = 0;
}


L4_INLINE void init_iterator(buffer_info *info, char *iterator)
{
	(char*)iterator = (char*)info->buffer;
}

L4_INLINE void scan_buffer(buffer_info *info, char *iterator)
{
	if(iterator + L4sample_size(iterator) + 5 < info->buffer)
		iterator += L4sample_size(iterator) + 5;
}

/***********************************************************************
 *	Functions for the alternative implementation					   *
 ***********************************************************************/

void min_time_unit(buffer_info *info, int *curr_unit, L4_Word_t tindex);
void print_matrix(void);
int window_factor(buffer_info *info);
void initialize_buffer(interrupt_spec *int_info, buffer_info *info, char *type_name, void (*fn)(void*));
void initialize_binfo(buffer_info *info);
void collect_samples(int sampling_params, interrupt_spec *int_info, buffer_info *info, void (*fn)(void*));
void interrupt_finished(int time_unit);
void print_intr_info(buffer_info *thread_buffer);
void write_tuple_output(interrupt_spec *int_info, buffer_info *info, void *data, char *sequence, L4_ThreadId_t tid_write);
int tuple_memory_size(void *data, char *sequence);
int memory_single_chunk(char *sequence);
L4_Word_t window_constant(buffer_info *info);
char *get_record(char *buff_ptr, int size);

void new_prof_thread(void (*fn)(void*));
L4_ThreadId_t create_temp_id(L4_Word_t raw_id);

void suspend(void (*fn)(void*));
void resume(void (*fn)(void*));
void kill(void (*fn)(void*));
#endif

